from collections import namedtuple
from typing import Optional, Sequence, Tuple, Union

import numpy as np
import SimpleITK as sitk

from imgtools.modules import Segmentation

INTERPOLATORS = {
    "linear": sitk.sitkLinear,
    "nearest": sitk.sitkNearestNeighbor,
    "bspline": sitk.sitkBSpline,
}


def resample(
    image: sitk.Image,
    spacing: Union[float, Sequence[float], np.ndarray],
    interpolation: str = "linear",
    anti_alias: bool = True,
    anti_alias_sigma: Optional[float] = None,
    transform: Optional[sitk.Transform] = None,
    output_size: Optional[Sequence[float]] = None,
) -> sitk.Image:
    """Resample image to a given spacing, optionally applying a transformation.

    Parameters
    ----------
    image
        The image to be resampled.

    spacing
        The new image spacing. If float, assumes the same spacing in all
        directions. Alternatively, a sequence of floats can be passed to
        specify spacing along each dimension. Passing 0 at any position will
        keep the original spacing along that dimension (useful for in-plane
        resampling).

    interpolation, optional
        The interpolation method to use. Valid options are:
        - "linear" for bi/trilinear interpolation (default)
        - "nearest" for nearest neighbour interpolation
        - "bspline" for order-3 b-spline interpolation

    anti_alias, optional
        Whether to smooth the image with a Gaussian kernel before resampling.
        Only used when downsampling, i.e. when `spacing < image.GetSpacing()`.
        This should be used to avoid aliasing artifacts.

    anti_alias_sigma, optional
        The standard deviation of the Gaussian kernel used for anti-aliasing.

    transform, optional
        Transform to apply to input coordinates when resampling. If None,
        defaults to identity transformation.

    output_size, optional
        Size of the output image. If None, it is computed to preserve the
        whole extent of the input image.

    Returns
    -------
    sitk.Image
        The resampled image.
    """

    try:
        interpolator = INTERPOLATORS[interpolation]
    except KeyError as ke:
        msg = f"interpolator must be one of {list(INTERPOLATORS.keys())}, got {interpolation}."
        raise ValueError(msg) from ke

    original_spacing = np.array(image.GetSpacing())
    original_size = np.array(image.GetSize())

    if isinstance(spacing, (float, int)):
        new_spacing = np.repeat(spacing, len(original_spacing)).astype(np.float64)
    else:
        spacing = np.asarray(spacing)
        new_spacing = np.where(spacing == 0, original_spacing, spacing)

    if output_size is None:
        new_size = np.round(original_size * original_spacing / new_spacing, decimals=0).astype(int)
    else:
        new_size = np.asarray(output_size)

    rif = sitk.ResampleImageFilter()
    rif.SetOutputOrigin(image.GetOrigin())
    rif.SetOutputSpacing(new_spacing)
    rif.SetOutputDirection(image.GetDirection())
    rif.SetSize(new_size.tolist())

    if transform is not None:
        rif.SetTransform(transform)

    downsample = new_spacing > original_spacing
    if downsample.any() and anti_alias:
        if not anti_alias_sigma:
            # sigma computation adapted from scikit-image
            # https://github.com/scikit-image/scikit-image/blob/master/skimage/transform/_warps.py
            anti_alias_sigma = np.maximum(1e-11, (original_spacing / new_spacing - 1) / 2)
        sigma = np.where(downsample, anti_alias_sigma, 1e-11)
        image = sitk.SmoothingRecursiveGaussian(image, sigma)

    rif.SetInterpolator(interpolator)
    resampled_image = rif.Execute(image)

    return resampled_image


def resize(
    image: sitk.Image,
    size: Union[int, Sequence[int], np.ndarray],
    interpolation: str = "linear",
    anti_alias: bool = True,
    anti_alias_sigma: Optional[float] = None,
) -> sitk.Image:
    """Resize image to a given size by resampling coordinates.

    Parameters
    ----------
    image
        The image to be resize.

    size
        The new image size. If float, assumes the same size in all directions.
        Alternatively, a sequence of floats can be passed to specify size along
        each dimension. Passing 0 at any position will keep the original
        size along that dimension.

    interpolation, optional
        The interpolation method to use. Valid options are:
        - "linear" for bi/trilinear interpolation (default)
        - "nearest" for nearest neighbour interpolation
        - "bspline" for order-3 b-spline interpolation

    anti_alias, optional
        Whether to smooth the image with a Gaussian kernel before resampling.
        Only used when downsampling, i.e. when `size < image.GetSize()`.
        This should be used to avoid aliasing artifacts.

    anti_alias_sigma, optional
        The standard deviation of the Gaussian kernel used for anti-aliasing.

    Returns
    -------
    sitk.Image
        The resized image.
    """

    original_size = np.array(image.GetSize())
    original_spacing = np.array(image.GetSpacing())

    if isinstance(size, (float, int)):
        new_size = np.repeat(size, len(original_size)).astype(np.float64)
    else:
        size = np.asarray(size)
        new_size = np.where(size == 0, original_size, size)

    new_spacing = original_spacing * original_size / new_size

    return resample(
        image,
        new_spacing,
        anti_alias=anti_alias,
        anti_alias_sigma=anti_alias_sigma,
        interpolation=interpolation,
        output_size=new_size.astype(int),
    )


def zoom(
    image: sitk.Image,
    scale_factor: Union[float, Sequence[float]],
    interpolation: str = "linear",
    anti_alias: bool = True,
    anti_alias_sigma: Optional[float] = None,
) -> sitk.Image:
    """Rescale image, preserving its spatial extent.

    The rescaled image will have the same spatial extent (size) but will be
    rescaled by `scale_factor` in each dimension. Alternatively, a separate
    scale factor for each dimension can be specified by passing a sequence
    of floats.

    Parameters
    ----------
    image
        The image to rescale.

    scale_factor
        If float, each dimension will be scaled by that factor. If tuple, each
        dimension will be scaled by the corresponding element.

    interpolation, optional
        The interpolation method to use. Valid options are:
        - "linear" for bi/trilinear interpolation (default)
        - "nearest" for nearest neighbour interpolation
        - "bspline" for order-3 b-spline interpolation

    anti_alias, optional
        Whether to smooth the image with a Gaussian kernel before resampling.
        Only used when downsampling, i.e. when `size < image.GetSize()`.
        This should be used to avoid aliasing artifacts.

    anti_alias_sigma, optional
        The standard deviation of the Gaussian kernel used for anti-aliasing.

    Returns
    -------
    sitk.Image
        The rescaled image.
    """
    dimension = image.GetDimension()

    if isinstance(scale_factor, float):
        scale_factor = (scale_factor,) * dimension

    centre_idx = np.array(image.GetSize()) / 2
    centre = image.TransformContinuousIndexToPhysicalPoint(centre_idx)

    transform = sitk.ScaleTransform(dimension, scale_factor)
    transform.SetCenter(centre)

    return resample(
        image,
        spacing=image.GetSpacing(),
        interpolation=interpolation,
        anti_alias=anti_alias,
        anti_alias_sigma=anti_alias_sigma,
        transform=transform,
        output_size=image.GetSize(),
    )


def rotate(
    image: sitk.Image,
    rotation_centre: Sequence[float],
    angles: Union[float, Sequence[float]],
    interpolation: str = "linear",
) -> sitk.Image:
    """Rotate an image around a given centre.

    Parameters
    ----------
    image
        The image to rotate.

    rotation_centre
        The centre of rotation in image coordinates.

    angles
        The angles of rotation around x, y and z axes.

    Returns
    -------
    sitk.Image
        The rotated image.
    """
    if isinstance(rotation_centre, np.ndarray):
        rotation_centre = rotation_centre.tolist()

    rotation_centre = image.TransformIndexToPhysicalPoint(rotation_centre)

    if image.GetDimension() == 2:
        rotation = sitk.Euler2DTransform(
            rotation_centre,
            angles,
            (0.0, 0.0),  # no translation
        )
    elif image.GetDimension() == 3:
        x_angle, y_angle, z_angle = angles

        rotation = sitk.Euler3DTransform(
            rotation_centre,
            x_angle,  # the angle of rotation around the x-axis, in radians -> coronal rotation
            y_angle,  # the angle of rotation around the y-axis, in radians -> saggittal rotation
            z_angle,  # the angle of rotation around the z-axis, in radians -> axial rotation
            (0.0, 0.0, 0.0),  # no translation
        )
    return resample(
        image, spacing=image.GetSpacing(), interpolation=interpolation, transform=rotation
    )


def crop(
    image: sitk.Image, crop_centre: Sequence[float], size: Union[int, Sequence[int], np.ndarray]
) -> sitk.Image:
    """Crop an image to the desired size around a given centre.

    Note that the cropped image might be smaller than size in a particular
    direction if the cropping window exceeds image boundaries.

    Parameters
    ----------
    image
        The image to crop.

    crop_centre
        The centre of the cropping window in image coordinates.

    size
        The size of the cropping window along each dimension in pixels. If
        float, assumes the same size in all directions. Alternatively, a
        sequence of floats can be passed to specify size along x, y and z
        dimensions. Passing 0 at any position will keep the original size along
        that dimension.

    Returns
    -------
    sitk.Image
        The cropped image.
    """
    crop_centre = np.asarray(crop_centre, dtype=np.float64)
    original_size = np.asarray(image.GetSize())

    size = np.array([size for _ in image.GetSize()]) if isinstance(size, int) else np.asarray(size)

    if (crop_centre < 0).any() or (crop_centre > original_size).any():
        msg = f"Crop centre outside image boundaries. Image size = {original_size}, crop centre = {crop_centre}"
        raise ValueError(msg)

    min_coords = np.clip(np.floor(crop_centre - size / 2).astype(np.int64), 0, original_size)
    min_coords = np.where(size == 0, 0, min_coords)

    max_coords = np.clip(np.floor(crop_centre + size / 2).astype(np.int64), 0, original_size)
    max_coords = np.where(size == 0, original_size, max_coords)

    min_x, min_y, min_z = min_coords
    max_x, max_y, max_z = max_coords

    return image[min_x:max_x, min_y:max_y, min_z:max_z]


# def constant_pad(image, size, cval=0.):
#     if isinstance(size, int):
#         size_lower = size_upper = [size for _ in image.GetSize()]
#     elif isinstance(size, (tuple, list, np.ndarray)):
#         if isinstance(size[0], int):
#             size_lower = size_upper = size
#         elif isinstance(size[0], (tuple, list, np.ndarray)):
#             size_lower = [s[0] for s in size]
#             size_upper = [s[1] for s in size]
#     else:
#         raise ValueError(
#             f"Size must be either int, sequence of int or sequence of sequences of ints, got {size}."
#         )
#     return sitk.ConstantPad(image, size_lower, size_upper, cval)

# def centre_on_point(image, centre):
#     pass

# def resize_by_cropping_or_padding(image, size, centre=None, cval=0.):
#     original_size = np.array(image.GetSize())
#     size = np.asarray(size)
#     centre = np.asarray(centre) if centre is not None else original_size / 2 # XXX is there any benefit to not using floor div here?

#     crop_dims = np.where(size < original_size)


def bounding_box(mask: sitk.Image, label: int = 1) -> Tuple[Tuple, Tuple]:
    """Find the axis-aligned bounding box of a region descriibed by a
    segmentation mask.

    Parameters
    ----------
    mask
        Segmentation mask describing the region of interest. Can be an image of
        type unsigned int representing a label map or `segmentation.Segmentation`.

    label, optional
        Label to use when computing bounding box if segmentation mask contains
        more than 1 labelled region.

    Returns
    -------
    tuple of tuples
        The bounding box location and size. The first tuple gives the
        coordinates of the corner closest to the origin and the second
        gives the size in pixels along each dimension.
    """

    if isinstance(mask, Segmentation):
        seg = Segmentation(mask)
        mask = seg.get_label(label=label, relabel=True)

    filter_ = sitk.LabelShapeStatisticsImageFilter()
    filter_.Execute(mask)
    bbox = filter_.GetBoundingBox(label)
    location = bbox[: len(bbox) // 2]
    size = bbox[len(bbox) // 2 :]
    return location, size


def centroid(mask: sitk.Image, label: int = 1, world_coordinates: bool = False) -> tuple:
    """Find the centroid of a labelled region specified by a segmentation mask.

    Parameters
    ----------
    mask
        Segmentation mask describing the region of interest. Can be an image of
        type unsigned int representing a label map or `segmentation.Segmentation`.

    label, optional
        Label to use when computing the centroid if segmentation mask contains
        more than 1 labelled region.

    world_coordinates, optional
        If True, return centroid in world coordinates, otherwise in image
        (voxel) coordinates (default).

    Returns
    -------
    tuple
        The centroid coordinates.
    """

    if isinstance(mask, Segmentation):
        seg = Segmentation(mask)
        mask = seg.get_label(label=label, relabel=True)

    filter_ = sitk.LabelShapeStatisticsImageFilter()
    filter_.Execute(mask)
    centroid_coords = filter_.GetCentroid(label)
    if not world_coordinates:
        centroid_coords = mask.TransformPhysicalPointToIndex(centroid_coords)
    return centroid_coords


def crop_to_mask_bounding_box(
    image: sitk.Image,
    mask: sitk.Image,
    margin: Union[int, Sequence[int], np.ndarray] = 0,
    label: int = 1,
) -> Tuple[sitk.Image]:
    """Crop the image using the bounding box of a region of interest specified
    by a segmentation mask.

    Parameters
    ----------
    image
        The image to crop.

    mask
        Segmentation mask describing the region of interest. Can be an image of
        type unsigned int representing a label map or `segmentation.Segmentation`.

    margin, optional
        A margin that will be added to each dimension when cropping. If int,
        add the same margin to each dimension. A sequence of ints can also be
        passed to specify the margin separately along each dimension.

    label, optional
        Label to use when computing the centroid if segmentation mask contains
        more than 1 labelled region.

    Returns
    -------
    tuple of sitk.Image
        The cropped image and mask.
    """

    if isinstance(mask, Segmentation):
        seg = Segmentation(mask)
        mask = seg.get_label(label=label, relabel=True)

    if isinstance(margin, Sequence):
        margin = np.asarray(margin)

    bbox_location, bbox_size = bounding_box(mask, label=label)
    bbox_location, bbox_size = np.array(bbox_location), np.array(bbox_size)
    crop_size = bbox_size + margin * 2
    crop_centre = bbox_location - margin + crop_size / 2

    image = crop(image, crop_centre, crop_size)
    mask = crop(mask, crop_centre, crop_size)

    return image, mask, crop_centre


def clip_intensity(image: sitk.Image, lower: float, upper: float) -> sitk.Image:
    """Clip image grey level intensities to specified range.

    The grey level intensities in the resulting image will fall in the range
    [lower, upper].

    Parameters
    ----------
    image
        The intensity image to clip.

    lower
        The lower bound on grey level intensity. Voxels with lower intensity
        will be set to this value.

    upper
        The upper bound on grey level intensity. Voxels with higer intensity
        will be set to this value.

    Returns
    -------
    sitk.Image
        The clipped intensity image.
    """
    return sitk.Clamp(image, image.GetPixelID(), lower, upper)


def window_intensity(image: sitk.Image, window: float, level: float) -> sitk.Image:
    """Restrict image grey level intensities to a given window and level.

    The grey level intensities in the resulting image will fall in the range
    [level - window / 2, level + window / 2].

    Parameters
    ----------
    image
        The intensity image to window.

    window
        The width of the intensity window.

    level
        The mid-point of the intensity window.

    Returns
    -------
    sitk.Image
        The windowed intensity image.
    """
    lower = level - window / 2
    upper = level + window / 2
    return clip_intensity(image, lower, upper)


def image_statistics(image: sitk.Image, mask: Optional[sitk.Image] = None, label: int = 1) -> float:
    """Compute the intensity statistics of an image.

    Returns the minimum, maximum, sum, mean, variance and standard deviation
    of image intensities.
    This function also supports computing the statistics in a specific
    region of interest if `mask` and `label` are passed.

    Parameters
    ----------
    image
        The image used to compute the statistics.

    mask, optional
        Segmentation mask specifying a region of interest used in computation.
        Can be an image of type unsigned int representing a label map or
        `segmentation.Segmentation`. Only voxels falling within the ROI will
        be considered. If None, use the whole image.

    label, optional
        Label to use when computing the statistics if segmentation mask contains
        more than 1 labelled region.

    Returns
    -------
    collections.namedtuple
        The computed intensity statistics in the image or region.
    """

    ImageStatistics = namedtuple(
        "ImageStatistics", ["minimum", "maximum", "sum", "mean", "variance", "standard_deviation"]
    )

    if mask is not None:
        if isinstance(mask, Segmentation):
            seg = Segmentation(mask)
            mask = seg.get_label(label=label, relabel=True)

        filter_ = sitk.LabelStatisticsImageFilter()
        filter_.Execute(image, mask)
        result = ImageStatistics(
            minimum=filter_.GetMinimum(label),
            maximum=filter_.GetMaximum(label),
            sum=filter_.GetSum(label),
            mean=filter_.GetMean(label),
            variance=filter_.GetVariance(label),
            standard_deviation=filter_.GetSigma(label),
        )
    else:
        filter_ = sitk.StatisticsImageFilter()
        filter_.Execute(image)
        result = ImageStatistics(
            minimum=filter_.GetMinimum(),
            maximum=filter_.GetMaximum(),
            sum=filter_.GetSum(),
            mean=filter_.GetMean(),
            variance=filter_.GetVariance(),
            standard_deviation=filter_.GetSigma(),
        )

    return result


def standard_scale(
    image: sitk.Image,
    mask: Optional[sitk.Image] = None,
    rescale_mean: Optional[float] = None,
    rescale_std: Optional[float] = None,
    label: int = 1,
) -> sitk.Image:
    """Rescale image intensities by subtracting the mean and dividing by
       standard deviation.

    If `rescale_mean` and `rescale_std` are None, image mean and standard
    deviation will be used, i.e. the resulting image intensities will have
    0 mean and unit variance. Alternatively, a specific mean and standard
    deviation can be passed to e.g. standardize a whole dataset of images.
    If a segmentation mask is passed, only the voxels falling within the mask
    will be considered when computing the statistics. However, the whole image
    will still be normalized using the computed values.

    Parameters
    ----------
    image
        The image to rescale.

    mask, optional
        Segmentation mask specifying a region of interest used in computation.
        Can be an image of type unsigned int representing a label map or
        `segmentation.Segmentation`. Only voxels falling within the ROI will
        be considered. If None, use the whole image.

    rescale_mean, optional
        The mean intensity used in rescaling. If None, image mean will be used.

    rescale_std, optional
        The standard deviation used in rescaling. If None, image standard
        deviation will be used.

    label, optional
        Label to use when computing the mean and standard deviation if
        segmentation mask contains more than 1 labelled region.

    Returns
    -------
    sitk.Image
        The rescaled image.
    """
    if not rescale_mean or not rescale_std:
        statistics = image_statistics(image, mask, label)
        rescale_mean = statistics.mean
        rescale_std = statistics.standard_deviation
    return (image - rescale_mean) / rescale_std


def min_max_scale(image: sitk.Image, minimum: float = 0.0, maximum: float = 1.0) -> sitk.Image:
    """Rescale image intensities to a given minimum and maximum.

    Applies a linear transformation to image intensities such that the minimum
    and maximum intensity values in the resulting image are equal to minimum
    (default 0) and maximum (default 1) respectively.

    Parameters
    ----------
    image
        The image to rescale.

    minimum, optional
        The minimum intensity in the rescaled image.

    maximum, optional
        The maximum intensity in the rescaled image.

    Returns
    -------
    sitk.Image
        The rescaled image.
    """
    return sitk.RescaleIntensity(image, minimum, maximum)



import pathlib
import re
from typing import List

import click

from imgtools.dicom import find_dicoms as find_dicoms_util
from imgtools.logging import logger


def natural_sort_key(s: str) -> list:
	"""Sort strings in a natural order."""
	return [int(text) if text.isdigit() else text.lower() for text in re.split(r'(\d+)', str(s))]


@click.command()
@click.argument(
	'path',
	type=click.Path(
		exists=True,
		file_okay=False,
		dir_okay=True,
		readable=True,
		path_type=pathlib.Path,
		resolve_path=True,
	),
)
@click.argument(
	'search_input',
	nargs=-1,  # Allow multiple search inputs
	type=str,
	required=False,
)
@click.option(
	'-e',
	'--extension',
	default='dcm',
	show_default=True,
	help='File extension to look for.',
)
@click.option(
	'-c',
	'--count',
	is_flag=True,
	default=False,
	show_default=True,
	help='Whether to just print the count of files found. This is useful for scripts.',
)
@click.option(
	'-l',
	'--limit',
	default=None,
	type=int,
	show_default=True,
	help='The limit of results to return.',
)
@click.option(
	'-ch',
	'--check-header',
	is_flag=True,
	default=False,
	show_default=True,
	help='Whether to check DICOM header for "DICM" signature.',
)
@click.option(
	'-s',
	'--sorted',
	'sort_results',
	is_flag=True,
	default=False,
	show_default=True,
	help='Sort the results alphabetically.',
)
@click.help_option(
	'-h',
	'--help',
)
def find_dicoms(
	search_input: List[str],
	path: pathlib.Path,
	extension: str,
	check_header: bool,
	count: bool,
	limit: int,
	sort_results: bool,
) -> None:
	"""A tool to find DICOM files.

	PATH is the directory to search for DICOM files.

	SEARCH_INPUT is an optional list of regex patterns to filter the search results.

	"""
	logger.info('Searching for DICOM files.', args=locals())

	dicom_files = find_dicoms_util(
		directory=path,
		check_header=check_header,
		recursive=True,
		extension=extension,
		limit=limit,  # Pass limit parameter
		search_input=search_input,
	)

	if not dicom_files:
		warningmsg = f'No DICOM files found in {path}.'
		if not dicom_files:
			warningmsg += f' Search input "{search_input}" did not match any DICOM files.'
			warningmsg += ' Note: ALL search inputs must match to return a result.'
		logger.warning(
			warningmsg,
			directory=path,
			check_header=check_header,
			recursive=True,
			extension=extension,
			limit=limit,
			search_input=search_input,
		)
		return

	logger.info('DICOM find successful.', count=len(dicom_files))

	if count:
		click.echo(f'Number of DICOM files found: {len(dicom_files)}')
		return

	if sort_results:
		dicom_files = sorted(dicom_files, key=lambda p: natural_sort_key(str(p)))

	logger.info('Search complete.')

	for dicom_file in dicom_files:
		click.echo(dicom_file)



if __name__ == '__main__':
	find_dicoms()



"""Manage and manipulate segmentation masks with multi-label support.

This module provides the `Segmentation` class and associated utilities for working
with medical image segmentation masks.
It extends the functionality of `SimpleITK.Image` to include ROI-specific operations,
label management, and metadata tracking.

Classes
-------
Segmentation
    A specialized class for handling multi-label segmentation masks. Includes
    functionality for extracting individual labels, resolving overlaps, and
    integrating with DICOM SEG metadata.

Functions
---------
accepts_segmentations(f)
    A decorator to ensure functions working on images handle `Segmentation` objects
    correctly by preserving metadata and ROI labels.

map_over_labels(segmentation, f, include_background=False, return_segmentation=True, **kwargs)
    Applies a function to each label in a segmentation mask and combines the results,
    optionally returning a new `Segmentation` object.

Notes
-----
- The `Segmentation` class tracks metadata and ROI names, enabling easier management
  of multi-label segmentation masks.
- The `generate_sparse_mask` method resolves overlapping contours by taking the
  maximum label value for each voxel, ensuring a consistent sparse representation.
- Integration with DICOM SEG metadata is supported through the `from_dicom_seg`
  class method, which creates `Segmentation` objects from DICOM SEG files.

Examples
--------
# Creating a Segmentation object from a SimpleITK.Image
>>> seg = Segmentation(
...     image,
...     roi_indices={
...         "GTV": 1,
...         "PTV": 2,
...     },
... )

# Extracting an individual label
>>> gtv_mask = seg.get_label(name="GTV")

# Generating a sparse mask
>>> sparse_mask = seg.generate_sparse_mask(verbose=True)

# Applying a function to each label in the segmentation
>>> def compute_statistics(label_image):
>>>     return sitk.LabelStatisticsImageFilter().Execute(label_image)

>>> stats = map_over_labels(
...     segmentation=seg,
...     f=compute_statistics,
... )
"""

from __future__ import annotations

from functools import wraps
from typing import Any, Callable, List, Optional, Set, Tuple, Union

import numpy as np
import SimpleITK as sitk

from imgtools.logging import logger
from imgtools.utils import array_to_image, image_to_array

from .sparsemask import SparseMask


def accepts_segmentations(f: Callable) -> Callable:
    """A decorator that ensures functions can handle `Segmentation` objects correctly.

    If the input image is an instance of `Segmentation`, the decorator preserves
    the ROI indices and raw ROI names in the output.

    This is useful when using functions that process images without losing metadata
    for the Segmentation class.

    Parameters
    ----------
    f : Callable
        The function to wrap, which processes an image.

    Returns
    -------
    Callable
        A wrapped function that preserves `Segmentation` metadata if the input
        is a `Segmentation` object.

    Examples
    --------
    >>> @accepts_segmentations
    ... def some_processing_function(
    ...     img,
    ...     *args,
    ...     **kwargs,
    ... ):
    ...     return img  # Perform some operation on the image
    >>> segmentation = Segmentation(
    ...     image,
    ...     roi_indices={
    ...         "ROI1": 1,
    ...         "ROI2": 2,
    ...     },
    ... )
    >>> result = some_processing_function(segmentation)
    >>> isinstance(
    ...     result,
    ...     Segmentation,
    ... )
    True
    """

    @wraps(f)
    def wrapper(
        img: Union[sitk.Image, Segmentation],
        *args: Any,  # noqa
        **kwargs: Any,  # noqa
    ) -> Union[sitk.Image, Segmentation]:
        result = f(img, *args, **kwargs)
        if isinstance(img, Segmentation):
            result = sitk.Cast(result, sitk.sitkVectorUInt8)
            return Segmentation(
                result, roi_indices=img.roi_indices, raw_roi_names=img.raw_roi_names
            )
        return result

    return wrapper


def map_over_labels(
    segmentation: Segmentation,
    f: Callable[[sitk.Image], sitk.Image],
    include_background: bool = False,
    return_segmentation: bool = True,
    **kwargs: Any,  # noqa
) -> Union[List[sitk.Image], Segmentation]:
    """
    Applies a function to each label in a segmentation mask.

    This function iterates over all labels in the segmentation mask, applies
    the provided function to each label individually, and optionally combines
    the results into a new `Segmentation` object.

    Parameters
    ----------
    segmentation : Segmentation
        The segmentation object containing multiple ROI labels.
    f : Callable[[sitk.Image], sitk.Image]
        A function to apply to each label in the segmentation.
    include_background : bool, optional
        If True, includes the background label (label 0) in the operation.
        Default is False.
    return_segmentation : bool, optional
        If True, combines the results into a new `Segmentation` object.
        If False, returns a list of processed labels as `sitk.Image`. Default is True.
    **kwargs : Any
        Additional keyword arguments passed to the function `f`.

    Returns
    -------
    Union[List[sitk.Image], Segmentation]
        A new `Segmentation` object if `return_segmentation` is True,
        otherwise a list of `sitk.Image` objects for each label.

    Examples
    --------
    >>> def threshold(
    ...     label_img,
    ...     threshold=0.5,
    ... ):
    ...     return sitk.BinaryThreshold(
    ...         label_img,
    ...         lowerThreshold=threshold,
    ...     )
    >>> segmentation = Segmentation(
    ...     image,
    ...     roi_indices={
    ...         "ROI1": 1,
    ...         "ROI2": 2,
    ...     },
    ... )
    >>> result = map_over_labels(
    ...     segmentation,
    ...     threshold,
    ...     threshold=0.5,
    ... )
    >>> isinstance(
    ...     result,
    ...     Segmentation,
    ... )
    True
    """
    if include_background:
        labels = range(segmentation.num_labels + 1)
    else:
        labels = range(1, segmentation.num_labels + 1)

    res = [f(segmentation.get_label(label=label), **kwargs) for label in labels]

    if return_segmentation and isinstance(res[0], sitk.Image):
        res = [sitk.Cast(r, sitk.sitkUInt8) for r in res]
        return Segmentation(
            sitk.Compose(*res),
            roi_indices=segmentation.roi_indices,
            raw_roi_names=segmentation.raw_roi_names,
        )
    return res


class Segmentation(sitk.Image):
    def __init__(
        self,
        segmentation: sitk.Image,
        metadata: Optional[dict] = None,
        roi_indices: Optional[dict[str, int]] = None,
        existing_roi_indices: Optional[dict[str, int]] = None,
        raw_roi_names: Optional[dict[str, int]] = None,
        frame_groups: Optional[Any] = None,  # noqa
    ) -> None:
        """Initializes the Segmentation class

        Parameters
        ----------
        roi_indices
            Dictionary of {"ROI": label number}

        existing_roi_indices
            Dictionary of {"ROI": label number} of the existing ROIs

        raw_roi_names
            Dictionary of {"ROI": original contour names}

        frame_groups
            PerFrameFunctionalGroupsSequence (5200, 9230) DICOM metadata
        """
        if raw_roi_names is None:
            raw_roi_names = {}
        if metadata is None:
            metadata = {}
        super().__init__(segmentation)
        self.num_labels = self.GetNumberOfComponentsPerPixel()
        self.raw_roi_names = raw_roi_names
        self.metadata = metadata
        self.frame_groups = frame_groups

        if not roi_indices:
            self.roi_indices = {f"label_{i}": i for i in range(1, self.num_labels + 1)}
        else:
            self.roi_indices = roi_indices
            if 0 in self.roi_indices.values():
                self.roi_indices = {k: v + 1 for k, v in self.roi_indices.items()}

        if len(self.roi_indices) != self.num_labels:
            for i in range(1, self.num_labels + 1):
                if i not in self.roi_indices.values():
                    self.roi_indices[f"label_{i}"] = i

        self.existing_roi_indices = existing_roi_indices

    def get_label(
        self,
        label: Optional[int] = None,
        name: Optional[str] = None,
        relabel: bool = False,
    ) -> sitk.Image:
        """
        Get the label image for a given label or name.

        Parameters
        ----------
        label : Optional[int]
            The label index to retrieve. If None, the name parameter must be provided.
        name : Optional[str]
            The name of the region of interest to retrieve. If None, the label parameter must be provided.
        relabel : bool
            If True, relabel the output image with the given label.

        Returns
        -------
        sitk.Image
            The label image corresponding to the given label or name.

        Raises
        ------
        ValueError
            If both label and name are None.
        """
        if label is None and name is None:
            raise ValueError("Must pass either label or name.")

        if label is None:
            # Retrieve the label index from the name
            label = self.roi_indices[name]

        if label == 0:
            # Background is stored implicitly and needs to be computed
            label_arr = sitk.GetArrayViewFromImage(self)
            # Create a binary image where background is 1 and other regions are 0
            label_img = sitk.GetImageFromArray((label_arr.sum(-1) == 0).astype(np.uint8))
        else:
            # Retrieve the label image for the given label index
            label_img = sitk.VectorIndexSelectionCast(self, label - 1)
            if relabel:
                # Relabel the output image with the given label
                label_img *= label

        return label_img

    def to_label_image(self) -> sitk.Image:
        """
        Convert the segmentation object to a label image.

        This method handles overlapping labels by taking the argmax of all overlaps.

        Returns
        -------
        sitk.Image
            The label image with each voxel assigned the label of the ROI with the highest value.
        """
        arr, *_ = image_to_array(self)
        label_arr = np.where(arr.sum(-1) != 0, arr.argmax(-1) + 1, 0)
        label_img = array_to_image(label_arr, reference_image=self)
        return label_img

    def __getitem__(self, idx) -> Segmentation | any:  # noqa
        res = super().__getitem__(idx)
        match res:
            case sitk.Image:
                res = Segmentation(
                    res, roi_indices=self.roi_indices, raw_roi_names=self.raw_roi_names
                )
            case _:
                pass

        return res

    def __repr__(self) -> str:
        return f"<Segmentation with ROIs: {self.roi_indices!r}>"

    def generate_sparse_mask(self, verbose: bool = False) -> SparseMask:
        """
        Generate a sparse mask from the contours, taking the argmax of all overlaps

        Parameters
        ----------
        mask
            Segmentation object to build sparse mask from

        Returns
        -------
        SparseMask
            The sparse mask object.
        """
        mask_arr = np.transpose(sitk.GetArrayFromImage(self))
        for name in self.roi_indices:
            self.roi_indices[name] = self.existing_roi_indices[name]

        sparsemask_arr = np.zeros(mask_arr.shape[1:])

        if verbose:
            voxels_with_overlap = set()

        if len(mask_arr.shape) == 4:
            for i in range(mask_arr.shape[0]):
                slc = mask_arr[i, :, :, :]
                slc *= list(self.roi_indices.values())[
                    i
                ]  # everything is 0 or 1, so this is fine to convert filled voxels to label indices
                if verbose:
                    res = self._max_adder(sparsemask_arr, slc)
                    sparsemask_arr = res[0]
                    for e in res[1]:
                        voxels_with_overlap.add(e)
                else:
                    sparsemask_arr = np.fmax(sparsemask_arr, slc)  # elementwise maximum
        else:
            sparsemask_arr = mask_arr

        sparsemask = SparseMask(sparsemask_arr, self.roi_indices)

        if verbose and len(voxels_with_overlap) != 0:
            msg = f"{len(voxels_with_overlap)} voxels have overlapping contours."
            logger.warning(msg)
        return sparsemask

    def _max_adder(
        self, arr_1: np.ndarray, arr_2: np.ndarray
    ) -> Tuple[np.ndarray, Set[Tuple[int, int, int]]]:
        """
        Takes the maximum of two 3D arrays elementwise and returns the resulting array and a list of voxels that have overlapping contours in a set

        Parameters
        ----------
        arr_1
            First array to take maximum of
        arr_2
            Second array to take maximum of

        Returns
        -------
        Tuple[np.ndarray, Set[Tuple[int, int, int]]]
            The resulting array and a list of voxels that have overlapping contours in a set
        """
        res = np.zeros(arr_1.shape)
        overlaps = {}  # set of tuples of the coords that have overlap
        for i in range(arr_1.shape[0]):
            for j in range(arr_1.shape[1]):
                for k in range(arr_1.shape[2]):
                    if arr_1[i, j, k] != 0 and arr_2[i, j, k] != 0:
                        overlaps.add((i, j, k))
                    res[i, j, k] = max(arr_1[i, j, k], arr_2[i, j, k])
        return res, overlaps

    @classmethod
    def from_dicom_seg(cls, mask: sitk.Image, meta: Any) -> Segmentation:  # noqa
        # get duplicates
        label_counters = {i.SegmentLabel: 1 for i in meta.SegmentSequence}
        raw_roi_names = {}  # {i.SegmentLabel: i.SegmentNumber for n, i in meta.SegmentSequence}
        for _n, i in enumerate(meta.SegmentSequence):
            label = i.SegmentLabel
            num = i.SegmentNumber

            if label not in raw_roi_names:
                raw_roi_names[label] = num
            else:
                raw_roi_names[f"{label}_{label_counters[label]}"] = num
                label_counters[label] += 1

        frame_groups = meta.PerFrameFunctionalGroupsSequence
        return cls(mask, raw_roi_names=raw_roi_names, frame_groups=frame_groups)



import os
import pathlib
from urllib import request
from zipfile import ZipFile

import pytest

from imgtools.logging import logger


@pytest.fixture(scope='session')
def curr_path():
    return pathlib.Path(__file__).parent.parent.resolve().as_posix()


@pytest.fixture(scope='session')
def dataset_path(curr_path):
    quebec_path = pathlib.Path(curr_path, 'data', 'Head-Neck-PET-CT')

    if not (quebec_path.exists() and len(list(quebec_path.glob('*'))) == 2):
        quebec_path.mkdir(parents=True, exist_ok=True)

        # Download QC dataset
        logger.info('Downloading the test dataset...')
        quebec_data_url = (
            'https://github.com/bhklab/tcia_samples/blob/main/Head-Neck-PET-CT.zip?raw=true'
        )
        quebec_zip_path = pathlib.Path(quebec_path, 'Head-Neck-PET-CT.zip').as_posix()
        request.urlretrieve(quebec_data_url, quebec_zip_path)
        with ZipFile(quebec_zip_path, 'r') as zipfile:
            zipfile.extractall(quebec_path)
        os.remove(quebec_zip_path)
    else:
        logger.info('Data already downloaded...')

    output_path = pathlib.Path(curr_path, 'tests', 'temp').as_posix()
    quebec_path = quebec_path.as_posix()

    # Dataset name
    dataset_name = os.path.basename(quebec_path)
    imgtools_path = pathlib.Path(os.path.dirname(quebec_path), '.imgtools')

    # Defining paths for autopipeline and dataset component
    crawl_path = pathlib.Path(imgtools_path, f'imgtools_{dataset_name}.csv').as_posix()
    edge_path = pathlib.Path(imgtools_path, f'imgtools_{dataset_name}_edges.csv').as_posix()
    # json_path =  pathlib.Path(imgtools_path, f"imgtools_{dataset_name}.json").as_posix()  # noqa: F841

    yield quebec_path, output_path, crawl_path, edge_path


@pytest.fixture(scope='session')
def modalities_path(curr_path):
    qc_path = pathlib.Path(curr_path, 'data', 'Head-Neck-PET-CT', 'HN-CHUS-052')
    assert qc_path.exists(), 'Dataset not found'

    path = {}
    path['CT'] = pathlib.Path(
        qc_path, '08-27-1885-CA ORL FDG TEP POS TX-94629/3.000000-Merged-06362'
    ).as_posix()
    path['RTSTRUCT'] = pathlib.Path(
        qc_path,
        '08-27-1885-OrophCB.0OrophCBTRTID derived StudyInstanceUID.-94629/Pinnacle POI-41418',
    ).as_posix()
    path['RTDOSE'] = pathlib.Path(
        qc_path,
        '08-27-1885-OrophCB.0OrophCBTRTID derived StudyInstanceUID.-94629/11376',
    ).as_posix()
    path['PT'] = pathlib.Path(
        qc_path, '08-27-1885-CA ORL FDG TEP POS TX-94629/532790.000000-LOR-RAMLA-44600'
    ).as_posix()
    return path



from argparse import ArgumentParser


def parser():
    parser = ArgumentParser("imgtools Automatic Processing Pipeline.")

    # arguments
    parser.add_argument("input_directory", type=str,
                        help="Path to top-level directory of dataset.")

    parser.add_argument("output_directory", type=str,
                        help="Path to output directory to save processed images.")

    parser.add_argument("--modalities", type=str, default="CT",
                        help="List of desired modalities. Type as string for ex: RTSTRUCT,CT,RTDOSE")

    parser.add_argument("--visualize", default=False, action="store_true",
                        help="Whether to visualize the data graph")

    parser.add_argument("--spacing", nargs=3, type=float, default=(1., 1., 0.),
                        help="The resampled voxel spacing in  (x, y, z) directions.")

    parser.add_argument("--n_jobs", type=int, default=-1,
                        help="The number of parallel processes to use.")

    parser.add_argument("--show_progress", action="store_true",
                        help="Whether to print progress to standard output.")

    parser.add_argument("--warn_on_error", default=False, action="store_true",
                        help="Whether to warn on error.")

    parser.add_argument("--overwrite", default=False, action="store_true",
                        help="Whether to write output files even if existing output files exist.")
    
    parser.add_argument("--nnunet", default=False, action="store_true",
                        help="Whether to make the output conform to nnunet requirements.")

    parser.add_argument("--train_size", type=float, default=1.0,
                        help="The proportion of data to be used for training, as a decimal.")

    parser.add_argument("--random_state", type=int, default=42,
                        help="The random state to be used for the train-test-split.")

    parser.add_argument("--read_yaml_label_names", default=False, action="store_true",
                        help="Whether to read the label names from roi_names.yaml in the input directory.")

    parser.add_argument("--ignore_missing_regex", default=False, action="store_true",
                        help="Whether to ignore patients with no ROI regexes that match the given ones. Will throw an error on patients without matches if this is not set.")

    parser.add_argument("--roi_yaml_path", type=str, default="",
                        help="Path to the YAML file defining ROI regexes")

    parser.add_argument("--custom_train_test_split", default=False, action="store_true",
                        help="Whether to use a custom train-test-split, stored in custom_train_test_split.yaml in the input directory.")

    parser.add_argument("--nnunet_inference", default=False, action="store_true",
                        help="Whether to generate data for nnUNet inference.")
    
    parser.add_argument("--dataset_json_path", type=str, default="",
                        help="Path to the dataset.json file defining image modality indices for nnUNet inference.")

    parser.add_argument("--continue_processing", default=False, action="store_true",
                        help="Whether to continue processing a partially completed dataset.")
    
    parser.add_argument("--dry_run", default=False, action="store_true",
                        help="Make a dry run of the pipeline, only producing the edge table and dataset.csv.")

    parser.add_argument("--verbose", default=False, action="store_true",
                        help="Verbose output flag.")

    parser.add_argument("--update", default=False, action="store_true",
                        help="Update crawled index. In other words, process from scratch.")

    parser.add_argument("--roi_select_first", default=False, action="store_true",
                        help="Only select first matching regex/ROI name.")

    parser.add_argument("--roi_separate", default=False, action="store_true",
                        help="Process each matching regex/ROI into separate masks. Each matched mask will be saved as ROI_n. (n = index from list of regex/ROIs)")

    # parser.add_argument("--custom_train_test_split_path", type=str,
    #                     help="Path to the YAML file defining the custom train-test-split.")

    return parser.parse_known_args()[0]



import os
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from imgtools.dicom.sort.sort_method import FileAction, handle_file


@pytest.fixture
def temp_dir():
    with TemporaryDirectory() as tmpdir:
        yield Path(tmpdir)


def test_move_file(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    destination = temp_dir / 'destination.txt'
    source.write_text('Test content')

    handle_file(source, destination, FileAction.MOVE)

    assert not source.exists()
    assert destination.exists()
    assert destination.read_text() == 'Test content'


def test_copy_file(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    destination = temp_dir / 'destination.txt'
    source.write_text('Test content')

    handle_file(source, destination, FileAction.COPY)

    assert source.exists()
    assert destination.exists()
    assert destination.read_text() == 'Test content'


def test_create_symlink(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    symlink = temp_dir / 'symlink.txt'
    source.write_text('Test content')

    handle_file(source, symlink, FileAction.SYMLINK)

    assert symlink.exists()
    assert symlink.is_symlink()
    assert symlink.read_text() == 'Test content'


def test_create_hardlink(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    hardlink = temp_dir / 'hardlink.txt'
    source.write_text('Test content')

    handle_file(source, hardlink, FileAction.HARDLINK)

    assert hardlink.exists()
    assert hardlink.stat().st_ino == source.stat().st_ino


def test_invalid_action(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    destination = temp_dir / 'destination.txt'
    source.write_text('Test content')

    with pytest.raises(ValueError):
        handle_file(source, destination, 'invalid_action')


def test_source_does_not_exist(temp_dir) -> None:
    source = temp_dir / 'non_existent.txt'
    destination = temp_dir / 'destination.txt'

    with pytest.raises(FileNotFoundError):
        handle_file(source, destination, FileAction.MOVE)


def test_no_write_permission(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    destination = temp_dir / 'no_permission_dir/destination.txt'
    source.write_text('Test content')

    os.chmod(temp_dir, 0o400)  # Read-only permission

    with pytest.raises(PermissionError):
        handle_file(source, destination, FileAction.MOVE)

    os.chmod(temp_dir, 0o700)  # Restore permissions


def test_overwrite_file(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    destination = temp_dir / 'destination.txt'
    source.write_text('Test content')
    destination.write_text('Old content')

    handle_file(source, destination, FileAction.COPY, overwrite=True)

    assert destination.read_text() == 'Test content'


def test_no_overwrite_file(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    destination = temp_dir / 'destination.txt'
    source.write_text('Test content')
    destination.write_text('Old content')

    with pytest.raises(FileExistsError):
        handle_file(source, destination, FileAction.COPY, overwrite=False)


def test_create_symlink_error(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    source.write_text('Test content')
    symlink = temp_dir / 'symlink.txt'
    symlink.symlink_to(source)  # Create a symlink to the source file

    # Create a symlink loop
    source.unlink()
    source.symlink_to(symlink)

    with pytest.raises(RuntimeError):
        FileAction.SYMLINK.create_symlink(source, symlink)


def test_create_hardlink_error(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    hardlink = temp_dir / 'hardlink.txt'
    source.write_text('Test content')
    hardlink.mkdir()  # Create a directory to cause an error

    with pytest.raises(RuntimeError, match='Failed to create hard link'):
        FileAction.HARDLINK.create_hardlink(source, hardlink)


def test_no_write_permission_error(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    destination = temp_dir / 'no_permission_dir/destination.txt'
    source.write_text('Test content')

    os.chmod(temp_dir, 0o400)  # Read-only permission

    with pytest.raises(PermissionError):
        handle_file(source, destination, FileAction.MOVE)

    os.chmod(temp_dir, 0o700)  # Restore permissions


def test_parent_directory_creation_error(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    destination = temp_dir / 'no_permission_dir/destination.txt'
    source.write_text('Test content')

    os.chmod(temp_dir, 0o400)  # Read-only permission

    with pytest.raises(PermissionError):
        handle_file(source, destination, FileAction.COPY)

    os.chmod(temp_dir, 0o700)  # Restore permissions


def test_choices() -> None:
    assert FileAction.choices() == ['move', 'copy', 'symlink', 'hardlink']


def test_no_write_permissions_for_directory(temp_dir) -> None:
    source = temp_dir / 'source.txt'
    destination = temp_dir / 'no_permission_dir/destination.txt'
    source.write_text('Test content')
    destination.parent.mkdir(0o400, parents=True)  # Create a read-only parent directory

    with pytest.raises(
        PermissionError, match='Failed to create parent directory or no write permission'
    ):
        handle_file(source, destination, FileAction.MOVE)

    os.chmod(temp_dir, 0o700)  # Restore permissions



from __future__ import annotations

from typing import Any, Iterator, NamedTuple, Sequence, Union

import numpy as np
import SimpleITK as sitk


class ImageGeometry(NamedTuple):
    size: Sequence[int]
    origin: Sequence[float]
    direction: Sequence[float]
    spacing: Sequence[float]


# Goals for this module:
# - One Image Class to Rule Them All (no more juggling numpy & sitk)
# - consistent indexing (z, y, x) (maybe fancy indexing?)
# - can be created either from array or sitk.Image
# - implements all basic operators of sitk.Image
# - easy conversion to array or sitk
# - method to apply arbitrary sitk filter
# - nicer repr

# TODO:
# - better support for masks (e.g. what happens when we pass a one-hot encoded mask?)
# - (optional) specify indexing order when creating new Image (sitk or numpy)

# Type alias for Image or Array
ImageOrArray = Union[sitk.Image, np.ndarray]


class Image:
    def __init__(
        self,
        image: ImageOrArray,
        geometry: ImageGeometry | None = None,
        origin: Sequence[float] | None = None,
        direction: Sequence[float] | None = None,
        spacing: Sequence[float] | None = None,
    ) -> None:
        if isinstance(image, sitk.Image):
            self._image = image
        elif isinstance(image, np.ndarray):
            if geometry is None and any((origin is None, direction is None, spacing is None)):
                raise ValueError('If image is a Numpy array, either geometry must be specified.')

            if geometry is not None:
                _, origin, direction, spacing = geometry

            self._image = sitk.GetImageFromArray(image)
            self._image.SetOrigin(origin[::-1])
            direction = tuple(direction)
            self._image.SetDirection(direction[:-3] + direction[3:6] + direction[:3])
            self._image.SetSpacing(spacing[::-1])
        else:
            msg = f'image must be either numpy.ndarray or SimpleITK.Image, not {type(image)}.'
            raise TypeError(msg)

    @property
    def size(self):
        return self._image.GetSize()[::-1]

    @property
    def origin(self):
        return self._image.GetOrigin()[::-1]

    @property
    def direction(self):
        direction = self._image.GetDirection()
        direction = direction[:-3] + direction[3:6] + direction[:3]
        return direction

    @property
    def spacing(self):
        return self._image.GetSpacing()[::-1]

    @property
    def geometry(self):
        return ImageGeometry(
            size=self.size,
            origin=self.origin,
            direction=self.direction,
            spacing=self.spacing,
        )

    @property
    def ndim(self):
        return len(self.size)

    @property
    def dtype(self):
        return self._image.GetPixelIDType()

    def astype(self, new_type):
        return Image(sitk.Cast(self._image, new_type))

    def to_sitk_image(self):
        return self._image

    def to_numpy(
        self, return_geometry: bool = False, view: bool = False
    ) -> Union[np.ndarray, tuple[np.ndarray, ImageGeometry]]:
        if view:
            array = sitk.GetArrayViewFromImage(self._image)
        else:
            array = sitk.GetArrayFromImage(self._image)
        if return_geometry:
            return array, self.geometry
        return array

    def __getitem__(self, idx: int | slice) -> Union[Image, Any]:  # noqa
        if isinstance(idx, (int, slice)):
            idx = (idx,)
        if len(idx) < self.ndim:
            idx += (slice(None),) * (self.ndim - len(idx))

        idx = idx[::-1]  # SimpleITK uses (x, y, z) ordering internally

        value = self._image[idx]

        try:  # XXX there probably is a nicer way to check if value is a scalar
            return Image(value)
        except TypeError:
            return value

    def __setitem__(self, idx: int | slice, value: Any) -> None:  # noqa
        if isinstance(idx, (int, slice)):
            idx = (idx,)
        if len(idx) < self.ndim:
            idx += (slice(None),) * (self.ndim - len(idx))

        idx = idx[::-1]  # SimpleITK uses (x, y, z) ordering internally

        value = self._image[idx]

        try:  # XXX there probably is a nicer way to check if value is a scalar
            return Image(value)
        except TypeError:
            return value

    def apply_filter(self, sitk_filter: sitk.ImageFilter) -> Union[Image, Any]:  # noqa
        result = sitk_filter.Execute(self._image)
        if isinstance(result, sitk.Image):
            return Image(result)
        else:
            return result

    def __neg__(self) -> Image:
        return Image(-self._image)

    def __abs__(self) -> Image:
        return Image(abs(self._image))

    def __invert__(self) -> Image:
        return Image(~self._image)

    def __add__(self, other: Union[Image, Image]) -> Image:
        other_val = getattr(other, '_image', other)
        return Image(self._image + other_val)

    def __sub__(self, other: Union[Image, Image]) -> Image:
        other_val = getattr(other, '_image', other)
        return Image(self._image - other_val)

    def __mul__(self, other: Union[Image, Image]) -> Image:
        other_val = getattr(other, '_image', other)
        return Image(self._image * other_val)

    def __div__(self, other: Union[Image, Image]) -> Image:
        other_val = getattr(other, '_image', other)
        return Image(self._image / other_val)

    def __floordiv__(self, other: Union[Image, Image]) -> Image:
        other_val = getattr(other, '_image', other)
        return Image(self._image // other_val)

    def __pow__(self, other: Union[Image, Image]) -> Image:
        other_val = getattr(other, '_image', other)
        return Image(self._image**other_val)

    def __iadd__(self, other: Union[Image, Image]) -> Image:
        other_val = getattr(other, '_image', other)
        self._image += other_val

    def __isub__(self, other: Union[Image, Image]) -> Image:
        other_val = getattr(other, '_image', other)
        self._image -= other_val

    def __imul__(self, other: Union[Image, Image]) -> Image:
        other_val = getattr(other, '_image', other)
        self._image *= other_val

    def __idiv__(self, other: Union[Image, Image]) -> Image:
        other_val = getattr(other, '_image', other)
        self._image /= other_val

    def __ifloordiv__(self, other: Union[Image, Image]) -> Image:
        other_val = getattr(other, '_image', other)
        self._image //= other_val

    def __iter__(self) -> Iterator:
        pass

    def __repr__(self) -> str:
        return (
            f'Image(image={self._image}, origin={self.origin}, '
            f'spacing={self.spacing}, direction={self.direction})'
        )

    def __str__(self) -> str:
        return (
            f'origin = {self.origin}\n'
            f'spacing = {self.spacing}\n'
            f'direction = {self.direction}\n'
            f'values = \n{self.to_numpy(view=True)}'
        )



import re
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, Tuple

from imgtools.dicom.sort.exceptions import InvalidPatternError
from imgtools.dicom.sort.parser import PatternParser
from imgtools.logging import logger


# Define custom exceptions
class PatternResolverError(Exception):
    """Base exception for errors in pattern resolution."""

    pass


@dataclass
class PatternResolver:
    r"""Handles parsing and validating filename patterns.

    By default, this class uses the following pattern parser:

    >>> DEFAULT_PATTERN: re.Pattern = re.compile(r'%(\w+)|\{(\w+)\}')

    This will match placeholders of the form `{key}` or `%(key)s`.

    Example
    -------
    Given a filename format like `"{subject_id}_{date}/{disease}.txt"`, the pattern parser
    will extract the following keys:

    >>> pattern_resolver.keys
    {'subject_id', 'date', 'disease'}

    And the following formatted pattern:

    >>> pattern_resolver.formatted_pattern
    %(subject_id)s_%(date)s/%(disease)s.txt

    So you could resolve the pattern like this:

    >>> data_dict = {
    ...     'subject_id': 'JohnDoe',
    ...     'date': 'January-01-2025',
    ...     'disease': 'cancer',
    ... }

    >>> pattern_resolver.formatted_pattern % data_dict
    'JohnDoe_01-01-2025/cancer.txt'

    A more convenient way to resolve the pattern is to use the `resolve` method:
    >>> pattern_resolver.resolve(data_dict))
    'JohnDoe_01-01-2025/cancer.txt'
    """

    filename_format: str = field(init=True)

    DEFAULT_PATTERN: ClassVar[re.Pattern] = re.compile(r"%(\w+)|\{(\w+)\}")

    pattern_parser: PatternParser = field(init=False)
    formatted_pattern: str = field(init=False)
    keys: list[str] = field(init=False)

    def __init__(self, filename_format: str) -> None:
        self.filename_format = filename_format

        try:
            self.pattern_parser = PatternParser(
                self.filename_format, pattern_parser=self.DEFAULT_PATTERN
            )
            self.formatted_pattern, self.keys = (
                self.parse()
            )  # Validate the pattern by parsing it
        except InvalidPatternError as e:
            msg = f"Invalid filename format: {e}"
            raise PatternResolverError(msg) from e
        else:
            logger.debug("All keys are valid.", keys=self.keys)
            logger.debug(
                "Formatted Pattern valid.", formatted_pattern=self.formatted_pattern
            )

    def parse(self) -> Tuple[str, list[str]]:
        """
        Parse and validate the pattern.

        Returns
        -------
        Tuple[str, List[str]]
            The formatted pattern string and a list of extracted keys.

        Raises
        ------
        InvalidPatternError
            If the pattern contains no valid placeholders or is invalid.
        """
        if hasattr(self, "formatted_pattern") and hasattr(self, "keys"):
            return self.formatted_pattern, self.keys

        self.formatted_pattern, self.keys = self.pattern_parser.parse()
        return self.formatted_pattern, self.keys

    def resolve(self, context: Dict[str, Any]) -> str:
        """Resolve the pattern using the provided context dictionary.

        Parameters
        ----------
        context : Dict[str, Any]
            Dictionary containing key-value pairs to substitute in the pattern.

        Returns
        -------
        str
            The resolved pattern string with placeholders replaced by values.

        Raises
        ------
        PatternResolverError
            If a required key is missing from the context dictionary.
        """

        # simultaneously check for None values and validate the pattern
        if len(none_keys := [k for k, v in context.items() if v is None]) > 0:
            msg = "None is not a valid value for a placeholder in the pattern."
            msg += f" None keys: {none_keys}"
            raise PatternResolverError(msg)

        try:
            return self.formatted_pattern % context
        except KeyError as e:
            # key error will be raised if formatted_pattern contains a key not in context
            missing_keys = set(context.keys()) - set(self.keys)
            msg = f"Missing value for placeholder(s): {missing_keys}"
            msg += "\nPlease provide a value for this key in the `context` argument."
            msg += f" i.e `{self.__class__.__name__}.save(..., {e.args[0]}=value)`."
            raise PatternResolverError(msg) from e



from __future__ import annotations

from typing import Dict, Optional, TypeVar, Union

import numpy as np
import SimpleITK as sitk
from matplotlib import pyplot as plt
from pydicom import Dataset, dcmread

from imgtools.logging import logger

T = TypeVar("T")


def read_image(path: str) -> sitk.Image:  # noqa
    reader = sitk.ImageSeriesReader()
    dicom_names = reader.GetGDCMSeriesFileNames(path)
    reader.SetFileNames(dicom_names)
    reader.MetaDataDictionaryArrayUpdateOn()
    reader.LoadPrivateTagsOn()

    return reader.Execute()


class Dose(sitk.Image):
    def __init__(
        self, img_dose: sitk.Image, df: Dataset, metadata: Optional[Dict[str, T]] = None
    ) -> None:
        super().__init__(img_dose)
        self.img_dose = img_dose
        self.df = df
        if metadata:
            self.metadata = metadata
        else:
            self.metadata = {}

    @classmethod
    def from_dicom_rtdose(cls, path: str) -> Dose:
        """
        Reads the data and returns the data frame and the image dosage in SITK format
        """
        dose = sitk.ReadImage(path) if ".dcm" in path else read_image(path)

        # if 4D, make 3D
        if dose.GetDimension() == 4:
            dose = dose[:, :, :, 0]

        # Get the metadata
        df = dcmread(path)

        # Convert to SUV
        factor = float(df.DoseGridScaling)
        img_dose = sitk.Cast(dose, sitk.sitkFloat32)
        img_dose = img_dose * factor

        metadata = {}

        return cls(img_dose, df, metadata)

    def resample_dose(self, ct_scan: sitk.Image) -> sitk.Image:
        """
        Resamples the RTDOSE information so that it can be overlayed with CT scan. The beginning and end slices of the
        resampled RTDOSE scan might be empty due to the interpolation
        """
        resampled_dose = sitk.Resample(
            self.img_dose, ct_scan
        )  # , interpolator=sitk.sitkNearestNeighbor)
        return resampled_dose

    def show_overlay(self, ct_scan: sitk.Image, slice_number: int) -> plt.Figure:
        """
        For a given slice number, the function resamples RTDOSE scan and overlays on top of the CT scan and returns the figure of the
        overlay
        """
        resampled_dose = self.resample_dose(ct_scan)
        fig = plt.figure("Overlayed RTdose image", figsize=[15, 10])
        dose_arr = sitk.GetArrayFromImage(resampled_dose)
        plt.subplot(1, 3, 1)
        plt.imshow(dose_arr[slice_number, :, :])
        plt.subplot(1, 3, 2)
        ct_arr = sitk.GetArrayFromImage(ct_scan)
        plt.imshow(ct_arr[slice_number, :, :])
        plt.subplot(1, 3, 3)
        plt.imshow(ct_arr[slice_number, :, :], cmap=plt.cm.gray)
        plt.imshow(dose_arr[slice_number, :, :], cmap=plt.cm.hot, alpha=0.4)
        return fig

    def get_metadata(
        self,
    ) -> Dict[Union[str, int], Union[str, float, Dict[str, Union[str, float, list]]]]:
        """
        Forms Dose-Value Histogram (DVH) from DICOM metadata
        {
            dvh_type
            dose_type
            dose_units
            vol_units
            ROI_ID: {
                vol: different volume values for different dosage bins
                dose_bins: different dose bins
                max_dose: max dose value
                mean_dose : mean dose value
                min_dose: min dose value
                total_vol: total volume of the ROI
            }
        }
        """
        try:
            num_roi = len(self.df.DVHSequence)
            self.dvh = {}
            # These properties are uniform across all the ROIs
            self.dvh["dvh_type"] = self.df.DVHSequence[0].DVHType
            self.dvh["dose_units"] = self.df.DVHSequence[0].DoseUnits
            self.dvh["dose_type"] = self.df.DVHSequence[0].DoseType
            self.dvh["vol_units"] = self.df.DVHSequence[0].DVHVolumeUnits
            # ROI specific properties
            for i in range(num_roi):
                raw_data = np.array(self.df.DVHSequence[i].DVHData)
                n = len(raw_data)

                # ROI ID
                roi_reference = (
                    self.df.DVHSequence[i].DVHReferencedROISequence[0].ReferencedROINumber
                )

                # Make dictionary for each ROI ID
                self.dvh[roi_reference] = {}

                # DVH specifc properties
                doses_bin = np.cumsum(raw_data[0:n:2])
                vol = raw_data[1:n:2]
                self.dvh[roi_reference]["dose_bins"] = doses_bin.tolist()
                self.dvh[roi_reference]["vol"] = vol.tolist()

                # ROI specific properties
                tot_vol = np.sum(vol)
                non_zero_index = np.where(vol != 0)[0]
                min_dose = doses_bin[non_zero_index[0]]
                max_dose = doses_bin[non_zero_index[-1]]
                mean_dose = np.sum(doses_bin * (vol / np.sum(vol)))
                self.dvh[roi_reference]["max_dose"] = max_dose
                self.dvh[roi_reference]["mean_dose"] = mean_dose
                self.dvh[roi_reference]["min_dose"] = min_dose
                self.dvh[roi_reference]["total_vol"] = tot_vol
        except AttributeError:
            logger.warning("No DVH information present in the DICOM. Returning empty dictionary")
            self.dvh = {}

        return self.dvh



from .database_handler import DatabaseHandler  # noqa
from .database import DICOMDatabaseInterface, DICOMIndexer, DICOMInsertMixin

__all__ = ["DatabaseHandler", "DICOMDatabaseInterface", "DICOMIndexer", "DICOMInsertMixin"]



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, List, Optional

from pydicom import dcmread
from sqlalchemy import JSON, Column, ForeignKey, String, Table
from sqlalchemy.orm import relationship
from sqlalchemy.orm.decl_api import registry

from imgtools.logging import logger

if TYPE_CHECKING:
    from pathlib import Path

mapper_registry = registry()


def repr_mixin(cls) -> type:  # type: ignore  # noqa: ANN001
    """
    A decorator to add a customizable __repr__ method to SQLAlchemy ORM models.

    Includes scalar fields (e.g., columns) and prints the length of lists for relationships.
    """

    def __repr_method__(self) -> str:  # type: ignore  # noqa: N807, ANN001
        # Fetch attributes and handle lists separately
        attrs = {
            key: (len(value) if isinstance(value, list) else value)
            for key, value in vars(self).items()
            if not key.startswith("_") and isinstance(value, (str, int, float, list))
        }
        class_name = self.__class__.__name__
        fields = ",\n  ".join(f"{key}={value}" for key, value in attrs.items())
        return f"<{class_name}(\n  {fields}\n)>"

    cls.__repr__ = __repr_method__
    return cls


####################################################################################################
# PATIENT
####################################################################################################


# Patient Table
@repr_mixin
@mapper_registry.mapped
@dataclass
class Patient:
    __table__ = Table(
        "patients",
        mapper_registry.metadata,
        Column("PatientID", String, primary_key=True, index=True),  # PRIMARY KEY
    )

    # Metadata from DICOM
    PatientID: str = field(init=True)

    # Relationships
    studies: List[Study] = field(default_factory=list)

    __mapper_args__ = {  # type: ignore
        "properties": {
            "studies": relationship("Study", back_populates="patient"),
        }
    }

    @classmethod
    def from_metadata(cls, metadata: Dict[str, str]) -> Patient:
        """
        Create a Patient instance from metadata.

        Parameters
        ----------
        metadata : Dict[str, Optional[str]]
                        Patient metadata.

        Returns
        -------
        Patient
                        A Patient instance.
        """
        return cls(
            PatientID=metadata["PatientID"],
        )

    @property
    def num_studies(self) -> int:
        """
        Get the number of studies for the patient.

        Returns
        -------
        int
                        The number of studies.
        """
        return len(self.studies)

    @property
    def num_series(self) -> int:
        """
        Get the number of series for the patient.

        Returns
        -------
        int
                        The number of series.
        """
        return sum(len(study.series) for study in self.studies)

    @property
    def modalities(self) -> Dict[str, int]:
        """
        Get the list of modalities for the patient.

        Returns
        -------
        List[str]
                        The list of modalities.
        """
        modalities = [modality for study in self.studies for modality in study.modalities]
        return dict(Counter(modalities))

    @property
    def series(self) -> List[Series]:
        """
        Get the series for the patient.

        Returns
        -------
        List[Series]
                        The series.
        """
        return [series for study in self.studies for series in study.series]

    @property
    def rtstructs(self) -> List[Series]:
        """
        Get the RTSTRUCT series for the patient.

        Returns
        -------
        List[Series]
                        The RTSTRUCT series.
        """
        return [series for series in self.series if series.Modality == "RTSTRUCT"]


####################################################################################################
# STUDY
####################################################################################################


# Study Table
@repr_mixin
@mapper_registry.mapped
@dataclass
class Study:
    __table__ = Table(
        "studies",
        mapper_registry.metadata,
        Column("StudyInstanceUID", String, primary_key=True, index=True),  # PRIMARY KEY
        Column("PatientID", String, ForeignKey("patients.PatientID"), nullable=False),
    )

    # Metadata from DICOM
    StudyInstanceUID: str = field(init=True)  # PRIMARY KEY
    PatientID: str = field(init=True)  # FOREIGN KEY

    # Relationships
    patient: Patient = field(init=False)
    series: List[Series] = field(default_factory=list)

    __mapper_args__ = {  # type: ignore
        "properties": {
            "patient": relationship("Patient", back_populates="studies"),
            "series": relationship("Series", back_populates="study"),
        }
    }

    @classmethod
    def from_metadata(cls, metadata: Dict[str, str]) -> Study:
        """
        Create a Study instance from metadata.

        Parameters
        ----------
        metadata : Dict[str, Optional[str]]
                        Study metadata.

        Returns
        -------
        Study
                        A Study instance.
        """
        return cls(
            StudyInstanceUID=metadata["StudyInstanceUID"],
            PatientID=metadata["PatientID"],
        )

    @property
    def num_files(self) -> int:
        """
        Get the number of files in the study.

        Returns
        -------
        int
                        The number of files.
        """
        return sum(series.num_files for series in self.series)

    @property
    def modalities(self) -> Dict[str, int]:
        """
        Get the list of modalities in the study.

        Returns
        -------
        List[str]
                        The list of modalities.
        """

        modalities = [series.Modality for series in self.series]
        return dict(Counter(modalities))

    @property
    def unique_modalities(self) -> List[str]:
        """
        Get the unique modalities in the study.

        Returns
        -------
        List[str]
                        The unique modalities.
        """
        return list(self.modalities.keys())


####################################################################################################
# SERIES, RTSTRUCTSERIES, ...
####################################################################################################


@repr_mixin
@mapper_registry.mapped
@dataclass
class Series:
    __table__ = Table(
        "series",
        mapper_registry.metadata,
        Column("SeriesInstanceUID", String, primary_key=True, index=True),  # PRIMARY KEY
        Column("StudyInstanceUID", String, ForeignKey("studies.StudyInstanceUID"), nullable=False),
        Column("Modality", String, nullable=False),
        Column("_RTSTRUCT", JSON, nullable=True),
        Column("_MR", JSON, nullable=True),
    )

    # Metadata from DICOM
    SeriesInstanceUID: str = field(init=True)  # PRIMARY KEY
    StudyInstanceUID: str = field(init=True)  # FOREIGN KEY
    Modality: str = field(init=True)
    _RTSTRUCT: Optional[Dict[str, str]] = field(default_factory=dict)
    _MR: Optional[Dict[str, str]] = field(default_factory=dict)

    # Relationships
    study: Study = field(init=False)
    images: List[Image] = field(default_factory=list)

    __mapper_args__ = {  # type: ignore
        "properties": {
            "study": relationship("Study", back_populates="series"),
            "images": relationship("Image", back_populates="series"),
        },
    }

    @classmethod
    def from_metadata(cls, metadata: Dict[str, str], file_path: Path) -> Series:
        base_cls = cls(
            SeriesInstanceUID=metadata["SeriesInstanceUID"],
            StudyInstanceUID=metadata["StudyInstanceUID"],
            Modality=metadata["Modality"],
        )

        match metadata["Modality"]:
            case "RTSTRUCT":
                try:
                    ds = dcmread(file_path, stop_before_pixels=True)
                    logger.info(f"Reading RTSTRUCT file: {file_path}")
                    roi_names = [roi.ROIName for roi in ds.StructureSetROISequence]
                    rfrs = ds.ReferencedFrameOfReferenceSequence[0]
                    ref_series_seq = (
                        rfrs.RTReferencedStudySequence[0]
                        .RTReferencedSeriesSequence[0]
                        .SeriesInstanceUID
                    )
                    frame_of_reference = rfrs.FrameOfReferenceUID
                    base_cls._RTSTRUCT = {
                        "ROINames": ",".join(roi_names),
                        "RTReferencedSeriesUID": ref_series_seq,
                        "FrameOfReferenceUID": frame_of_reference,
                    }
                except Exception as e:
                    logger.exception(f"Error reading RTSTRUCT file: {file_path}", error=str(e))
                    raise e
            case "MR":
                try:
                    ds = dcmread(file_path, stop_before_pixels=True)
                    logger.info(f"Reading MR file: {file_path}")
                    _mr = {
                        "RepetitionTime": ds.RepetitionTime,
                        "EchoTime": ds.EchoTime,
                        "SliceThickness": ds.SliceThickness,
                        "ScanningSequence": ds.ScanningSequence,
                        "MagneticFieldStrength": ds.MagneticFieldStrength,
                        "ImagedNucleus": ds.ImagedNucleus,
                    }

                    base_cls._MR = {k: str(v) for k, v in _mr.items()}
                except Exception as e:
                    logger.exception(f"Error reading MR file: {file_path}", error=str(e))
                    raise e
            case _:
                pass

        return base_cls

    @property
    def num_files(self) -> int:
        return len(self.images)

    @property
    def RTReferencedSeriesUID(self) -> Optional[str]:  # noqa: N802
        return self._RTSTRUCT.get("RTReferencedSeriesUID", None) if self._RTSTRUCT else None

    @property
    def ROINames(self) -> Optional[str]:  # noqa: N802
        return self._RTSTRUCT.get("ROINames", None) if self._RTSTRUCT else None


####################################################################################################
# IMAGE
####################################################################################################


# Image Table
@repr_mixin
@mapper_registry.mapped
@dataclass
class Image:
    __table__ = Table(
        "images",
        mapper_registry.metadata,
        Column("FilePath", String, primary_key=True, index=True),  # PRIMARY KEY
        Column("SOPInstanceUID", String, nullable=False),
        Column("SeriesInstanceUID", String, ForeignKey("series.SeriesInstanceUID"), nullable=False),
    )

    # Metadata from DICOM
    FilePath: str = field(init=True)
    SOPInstanceUID: str = field(init=True)
    SeriesInstanceUID: str = field(init=True)

    # Relationships
    series: Series = field(init=False)

    __mapper_args__ = {  # type: ignore
        "properties": {
            "series": relationship("Series", back_populates="images"),
        }
    }

    @classmethod
    def from_metadata(cls, metadata: Dict[str, str], file_path: Path) -> Image:
        """
        Create a Image instance from metadata.

        Parameters
        ----------
        metadata : Dict[str, Optional[str]]
                        Image metadata.

        Returns
        -------
        Image
                        A Image instance.
        """
        return cls(
            SOPInstanceUID=metadata["SOPInstanceUID"],
            SeriesInstanceUID=metadata["SeriesInstanceUID"],
            FilePath=file_path.as_posix(),
        )



from __future__ import annotations

import contextlib
import datetime
import os
import pathlib
from typing import TYPE_CHECKING, Dict, Optional, TypeVar, Union

import numpy as np
import SimpleITK as sitk
from matplotlib import pyplot as plt
from pydicom import dcmread

from imgtools.logging import logger

if TYPE_CHECKING:
    from pydicom.dataset import FileDataset

T = TypeVar("T")


def read_image(path: str, series_id: Optional[str] = None) -> sitk.Image:
    reader = sitk.ImageSeriesReader()
    dicom_names = reader.GetGDCMSeriesFileNames(path, seriesID=series_id if series_id else "")
    reader.SetFileNames(dicom_names)
    reader.MetaDataDictionaryArrayUpdateOn()
    reader.LoadPrivateTagsOn()

    return reader.Execute()


class PET(sitk.Image):
    def __init__(
        self,
        img_pet: sitk.Image,
        df: FileDataset,
        factor: float,
        calc: bool,
        metadata: Optional[Dict[str, Union[str, float, bool]]] = None,
    ) -> None:
        super().__init__(img_pet)
        self.img_pet: sitk.Image = img_pet
        self.df: FileDataset = df
        self.factor: float = factor
        self.calc: bool = calc
        self.metadata: Dict[str, Union[str, float, bool]] = metadata if metadata else {}

    @classmethod
    def from_dicom_pet(
        cls, path: str, series_id: Optional[str] = None, pet_image_type: str = "SUV"
    ) -> PET:
        """Read the PET scan and returns the data frame and the image dosage in SITK format

        There are two types of existing formats which has to be mentioned in the type
        type:
            SUV: gets the image with each pixel value having SUV value
            ACT: gets the image with each pixel value having activity concentration
        SUV = Activity concenteration/(Injected dose quantity/Body weight)

        Please refer to the pseudocode: https://qibawiki.rsna.org/index.php/Standardized_Uptake_Value_(SUV)
        If there is no data on SUV/ACT then backup calculation is done based on the formula in the documentation, although, it may
        have some error.
        """
        pet: sitk.Image = read_image(path, series_id)
        path_one: str = pathlib.Path(path, os.listdir(path)[0]).as_posix()
        df: FileDataset = dcmread(path_one)
        calc: bool = False
        try:
            if pet_image_type == "SUV":
                factor: float = df.to_json_dict()["70531000"]["Value"][0]
            else:
                factor: float = df.to_json_dict()["70531009"]["Value"][0]
        except KeyError:
            logger.warning(
                "Scale factor not available in DICOMs. Calculating based on metadata, may contain errors"
            )
            factor = cls.calc_factor(df, pet_image_type)
            calc = True
        img_pet: sitk.Image = sitk.Cast(pet, sitk.sitkFloat32)

        # SimpleITK reads some pixel values as negative but with correct value
        img_pet = sitk.Abs(img_pet * factor)

        metadata: Dict[str, Union[str, float, bool]] = {}
        return cls(img_pet, df, factor, calc, metadata)

    def get_metadata(self) -> Dict[str, Union[str, float, bool]]:
        self.metadata = {}
        with contextlib.suppress(Exception):
            self.metadata["weight"] = float(self.df.PatientWeight)
        try:
            self.metadata["scan_time"] = datetime.datetime.strptime(
                self.df.AcquisitionTime, "%H%M%S.%f"
            )
            self.metadata["injection_time"] = datetime.datetime.strptime(
                self.df.RadiopharmaceuticalInformationSequence[0].RadiopharmaceuticalStartTime,
                "%H%M%S.%f",
            )
            self.metadata["half_life"] = float(
                self.df.RadiopharmaceuticalInformationSequence[0].RadionuclideHalfLife
            )
            self.metadata["injected_dose"] = float(
                self.df.RadiopharmaceuticalInformationSequence[0].RadionuclideTotalDose
            )
        except KeyError:
            pass
        self.metadata["factor"] = self.factor
        self.metadata["Values_Assumed"] = self.calc
        return self.metadata

    def resample_pet(self, ct_scan: sitk.Image) -> sitk.Image:
        resampled_pt: sitk.Image = sitk.Resample(self.img_pet, ct_scan)
        return resampled_pt

    def show_overlay(self, ct_scan: sitk.Image, slice_number: int) -> plt.figure:
        resampled_pt: sitk.Image = self.resample_pet(ct_scan)
        fig: plt.figure = plt.figure("Overlayed image", figsize=[15, 10])
        pt_arr: np.ndarray = sitk.GetArrayFromImage(resampled_pt)
        plt.subplot(1, 3, 1)
        plt.imshow(pt_arr[slice_number, :, :])
        plt.subplot(1, 3, 2)
        ct_arr: np.ndarray = sitk.GetArrayFromImage(ct_scan)
        plt.imshow(ct_arr[slice_number, :, :])
        plt.subplot(1, 3, 3)
        plt.imshow(ct_arr[slice_number, :, :], cmap=plt.cm.gray)
        plt.imshow(pt_arr[slice_number, :, :], cmap=plt.cm.hot, alpha=0.4)
        return fig

    @staticmethod
    def calc_factor(df: FileDataset, pet_image_type: str) -> float:
        try:
            weight: float = float(df.PatientWeight) * 1000
        except KeyError:
            logger.warning("Patient Weight Not Present. Taking 75Kg")
            weight = 75000
        try:
            scan_time: datetime.datetime = datetime.datetime.strptime(
                df.AcquisitionTime, "%H%M%S.%f"
            )
            injection_time: datetime.datetime = datetime.datetime.strptime(
                df.RadiopharmaceuticalInformationSequence[0].RadiopharmaceuticalStartTime,
                "%H%M%S.%f",
            )
            half_life: float = float(
                df.RadiopharmaceuticalInformationSequence[0].RadionuclideHalfLife
            )
            injected_dose: float = float(
                df.RadiopharmaceuticalInformationSequence[0].RadionuclideTotalDose
            )

            a: float = np.exp(-np.log(2) * ((scan_time - injection_time).seconds / half_life))

            injected_dose_decay: float = a * injected_dose
        except Exception:
            logger.warning("Not enough data available, taking average values")
            a = np.exp(-np.log(2) * (1.75 * 3600) / 6588)
            injected_dose_decay = 420000000 * a

        suv: float = weight / injected_dose_decay
        if pet_image_type == "SUV":
            return suv
        else:
            return 1 / a



from contextlib import contextmanager
from pathlib import Path
from typing import Generator

from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker

from imgtools.dicom.index.models import mapper_registry
from imgtools.logging import logger


class DatabaseHandler:
    """
    Manages database operations using SQLAlchemy ORM.
    """

    def __init__(self, db_path: Path, force_delete: bool = False) -> None:
        """
        Initialize the DatabaseHandler with an SQLite database.

        Parameters
        ----------
        db_path : Path
                Path to the SQLite database file.
        force_delete : bool
                Whether to delete the existing database file if it exists.
        """
        if force_delete and db_path.exists():
            logger.debug("Deleting existing database file.", db_path=db_path)
            db_path.unlink()  # Delete the existing database file

        logger.debug("Creating database engine.", db_path=db_path)
        self.engine = create_engine(f"sqlite:///{db_path}", future=True)
        mapper_registry.metadata.create_all(self.engine)  # Create tables
        self.Session = sessionmaker(bind=self.engine)

    @contextmanager
    def session(self) -> Generator[Session, None, None]:
        """
        Create a new SQLAlchemy session.

        Yields
        ------
        Session
                A SQLAlchemy session object.
        """
        session = self.Session()
        try:
            yield session
            session.commit()  # Commit the transaction
        except Exception:
            session.rollback()  # Rollback on exception
            raise
        finally:
            session.close()



import click

from imgtools import __version__

from . import set_log_verbosity
from .dicomfind import find_dicoms
from .dicomsort import dicomsort


@click.group(
    no_args_is_help=True,
)
@set_log_verbosity()
@click.version_option(
    version=__version__,
    package_name="med-imagetools",
    prog_name="imgtools",
    message="%(package)s:%(prog)s:%(version)s",
)
@click.help_option(
    "-h",
    "--help",
)
def cli(verbose: int, quiet: bool) -> None:
    """A collection of tools for working with medical imaging data."""
    pass


cli.add_command(dicomsort)
cli.add_command(find_dicoms, "find-dicoms")

if __name__ == "__main__":
    cli()



"""
This file does not get tested because it doesnt start with test_ and so, is not a part of the test suite.

Leaving here for future reference.
"""

import os
import pathlib
import re
from typing import List
from urllib import request
from zipfile import ZipFile

import pandas as pd
import pytest
import torch
import torchio as tio

from imgtools.io import Dataset


@pytest.fixture(scope='session')
def dataset_path():
    curr_path = pathlib.Path(__file__).parent.parent.resolve()
    quebec_path = pathlib.Path(pathlib.Path(curr_path, 'data', 'Head-Neck-PET-CT').as_posix())

    if not os.path.exists(quebec_path):
        pathlib.Path(quebec_path).mkdir(parents=True, exist_ok=True)
        # Download QC dataset
        quebec_data_url = (
            'https://github.com/bhklab/tcia_samples/blob/main/Head-Neck-PET-CT.zip?raw=true'
        )
        quebec_zip_path = pathlib.Path(quebec_path, 'Head-Neck-PET-CT.zip').as_posix()
        request.urlretrieve(quebec_data_url, quebec_zip_path)
        with ZipFile(quebec_zip_path, 'r') as zipfile:
            zipfile.extractall(quebec_path)
        os.remove(quebec_zip_path)
    else:
        pass
    output_path = pathlib.Path(curr_path, 'tests', 'temp').as_posix()
    quebec_path = quebec_path.as_posix()

    # Dataset name
    dataset_name = os.path.basename(quebec_path)
    imgtools_path = pathlib.Path(os.path.dirname(quebec_path), '.imgtools')

    # Defining paths for autopipeline and dataset component
    crawl_path = pathlib.Path(imgtools_path, f'imgtools_{dataset_name}.csv').as_posix()
    json_path = pathlib.Path(imgtools_path, f'imgtools_{dataset_name}.json').as_posix()  # noqa: F841
    edge_path = pathlib.Path(imgtools_path, f'imgtools_{dataset_name}_edges.csv').as_posix()
    assert (
        os.path.exists(crawl_path) & os.path.exists(edge_path) & os.path.exists(json_path)
    ), 'There was no crawler output'

    yield quebec_path, output_path, crawl_path, edge_path


class select_roi_names(tio.LabelTransform):
    """
    Based on the given roi names, selects from the given set
    """

    def __init__(self, roi_names: List[str] = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.kwargs = kwargs
        self.roi_names = roi_names

    def apply_transform(self, subject):
        # list of roi_names
        for image in self.get_images(subject):
            # For only applying to labelmaps
            metadata = subject['metadata_RTSTRUCT_CT']
            patterns = self.roi_names
            mask = torch.empty_like(image.data)[: len(patterns)]
            for j, pat in enumerate(patterns):
                k = []
                for i, col in enumerate(metadata):
                    if re.match(pat, col, flags=re.IGNORECASE):
                        k.append(i)
                if len(k) == 0:
                    mask[j] = mask[j] * 0
                else:
                    mask[j] = (image.data[k].sum(axis=0) > 0) * 1
            image.set_data(mask)
        return subject

    def is_invertible(self) -> bool:
        return False


# Defining for test_dataset method in Test_components class
def collate_fn(data):
    """
    data: is a tio.subject with multiple columns
          Need to return required data
    """
    mod_names = [items for items in data[0] if items.split('_')[0] == 'mod']
    temp_stack = {}
    for names in mod_names:
        temp_stack[names] = torch.stack(tuple(items[names].data for items in data))
    return temp_stack


@pytest.mark.parametrize('modalities', ['CT', 'CT,RTSTRUCT', 'CT,RTSTRUCT,RTDOSE'])
class TestDataset:
    """
    For testing the dataset components of the med-imagetools package
    test_dataset:
        1) Checks if the length of the dataset matches
        2) Checks if the items in the subject object is correct and present
        3) Checks if you are able to load it via load_nrrd and load_directly, and checks if the subjects generated matches
        4) Checks if torch data loader can load the formed dataset and get atleast 1 iteration
        5) Checks if the transforms are happening by checking the size
    """

    @pytest.fixture(autouse=True)
    def _get_path(self, dataset_path) -> None:
        self.input_path, self.output_path, self.crawl_path, self.edge_path = dataset_path

    def test_dataset(self, modalities) -> None:
        """
        Testing the Dataset class
        """
        output_path_mod = pathlib.Path(
            self.output_path, str('temp_folder_' + ('_').join(modalities.split(',')))
        ).as_posix()
        comp_path = pathlib.Path(output_path_mod).resolve().joinpath('dataset.csv').as_posix()
        pd.read_csv(comp_path, index_col=0)

        # Loading from nrrd files
        subjects_nrrd = Dataset.load_image(output_path_mod, ignore_multi=True)
        # Loading files directly
        # subjects_direct = Dataset.load_directly(self.input_path,modalities=modalities,ignore_multi=True)

        # The number of subjects is equal to the number of components which is 2 for this dataset
        # assert len(subjects_nrrd) == len(subjects_direct) == 2, "There was some error in generation of subject object"
        # assert subjects_nrrd[0].keys() == subjects_direct[0].keys()

        # del subjects_direct
        # To check if all metadata items present in the keys
        # temp_nrrd = subjects_nrrd[0]
        # columns_shdbe_present = set([col if col.split("_")[0]=="metadata" else "mod_"+("_").join(col.split("_")[1:]) for col in list(comp_table.columns) if col.split("_")[0] in ["folder","metadata"]])
        # print(columns_shdbe_present)
        # assert set(temp_nrrd.keys()).issubset(columns_shdbe_present), "Not all items present in dictionary, some fault in going through the different columns in a single component"

        transforms = tio.Compose(
            [
                tio.Resample(4),
                tio.CropOrPad((96, 96, 40)),
                select_roi_names(['larynx']),
                tio.OneHot(),
            ]
        )

        # Forming dataset and dataloader
        test_set = tio.SubjectsDataset(subjects_nrrd, transform=transforms)
        test_loader = torch.utils.data.DataLoader(
            test_set, batch_size=2, shuffle=True, collate_fn=collate_fn
        )

        # Check test_set is correct
        assert len(test_set) == 2

        # Get items from test loader
        # If this function fails , there is some error in formation of test
        data = next(iter(test_loader))
        A = [
            item[1].shape == (2, 1, 96, 96, 40)
            if 'RTSTRUCT' not in item[0]
            else item[1].shape == (2, 2, 96, 96, 40)
            for item in data.items()
        ]
        assert all(A), 'There is some problem in the transformation/the formation of subject object'



import json
import os
import pathlib
from argparse import ArgumentParser

import pandas as pd
from joblib import Parallel, delayed
from pydicom import dcmread
from tqdm import tqdm

from imgtools.logging import logger


# fmt: off
def crawl_one(folder: pathlib.Path) -> dict:
    folder_path = pathlib.Path(folder)
    database = {}
    for path in folder_path.iterdir():
        # find dicoms
        dicoms = [f for f in path.rglob("*.dcm")]
        logger.info(f"Found {len(dicoms)} dicoms in {path}")

        for dcm in dicoms:
            try:
                dcm_path = pathlib.Path(dcm)

                fname = dcm_path.name

                rel_path = dcm_path.relative_to(
                    folder_path.parent.parent
                )  # rel_path of dicom from folder

                rel_posix = (
                    rel_path.parent.as_posix()
                )  # folder name + until parent folder of dicom

                meta = dcmread(dcm, force=True, stop_before_pixels=True)
                patient = str(meta.PatientID)
                study = str(meta.StudyInstanceUID)
                series = str(meta.SeriesInstanceUID)
                instance = str(meta.SOPInstanceUID)

                (
                    reference_ct,
                    reference_rs,
                    reference_pl,
                ) = "", "", ""
                tr, te, tesla, scan_seq, elem = "", "", "", "", ""
                try:
                    orientation = str(meta.ImageOrientationPatient)  # (0020, 0037)
                except:
                    orientation = ""

                try:
                    orientation_type = str(
                        meta.AnatomicalOrientationType
                    )  # (0010, 2210)
                except:
                    orientation_type = ""

                try:  # RTSTRUCT
                    reference_ct = str(
                        meta.ReferencedFrameOfReferenceSequence[0]
                        .RTReferencedStudySequence[0]
                        .RTReferencedSeriesSequence[0]
                        .SeriesInstanceUID
                    )
                except:
                    try:  # SEGMENTATION
                        reference_ct = str(
                            meta.ReferencedSeriesSequence[0].SeriesInstanceUID
                        )
                    except:
                        try:  # RTDOSE
                            reference_rs = str(
                                meta.ReferencedStructureSetSequence[
                                    0
                                ].ReferencedSOPInstanceUID
                            )
                        except:
                            pass
                        try:
                            reference_ct = str(
                                meta.ReferencedImageSequence[0].ReferencedSOPInstanceUID
                            )
                        except:
                            pass
                        try:
                            reference_pl = str(
                                meta.ReferencedRTPlanSequence[
                                    0
                                ].ReferencedSOPInstanceUID
                            )
                        except:
                            pass

                # MRI Tags
                try:
                    tr = float(meta.RepetitionTime)
                except:
                    pass
                try:
                    te = float(meta.EchoTime)
                except:
                    pass
                try:
                    scan_seq = str(meta.ScanningSequence)
                except:
                    pass
                try:
                    tesla = float(meta.MagneticFieldStrength)
                except:
                    pass
                try:
                    elem = str(meta.ImagedNucleus)
                except:
                    pass

                # Frame of Reference UIDs
                try:
                    reference_frame = str(meta.FrameOfReferenceUID)
                except:
                    try:
                        reference_frame = str(
                            meta.ReferencedFrameOfReferenceSequence[
                                0
                            ].FrameOfReferenceUID
                        )
                    except:
                        reference_frame = ""

                try:
                    study_description = str(meta.StudyDescription)
                except:
                    study_description = ""

                try:
                    series_description = str(meta.SeriesDescription)
                except:
                    series_description = ""

                try:
                    subseries = str(meta.AcquisitionNumber)
                except:
                    subseries = "default"

                if patient not in database:
                    database[patient] = {}
                if study not in database[patient]:
                    database[patient][study] = {"description": study_description}
                if series not in database[patient][study]:
                    rel_crawl_path = rel_posix
                    if meta.Modality == "RTSTRUCT":
                        rel_crawl_path = os.path.join(rel_crawl_path, fname)

                    database[patient][study][series] = {
                        "description": series_description
                    }
                if subseries not in database[patient][study][series]:
                    database[patient][study][series][subseries] = {'instances': {},
                                                                   'instance_uid': instance,
                                                                   'modality': meta.Modality,
                                                                   'reference_ct': reference_ct,
                                                                   'reference_rs': reference_rs,
                                                                   'reference_pl': reference_pl,
                                                                   'reference_frame': reference_frame,
                                                                   'folder': rel_crawl_path,
                                                                   'orientation': orientation,
                                                                   'orientation_type': orientation_type,
                                                                   'repetition_time':tr,
                                                                   'echo_time':te,
                                                                   'scan_sequence': scan_seq,
                                                                   'mag_field_strength': tesla,
                                                                   'imaged_nucleus': elem,
                                                                   'fname': rel_path.as_posix()  # temporary until we switch to json-based loading
                                                                   }
                database[patient][study][series][subseries]["instances"][instance] = (
                    rel_path.as_posix()
                )
            except Exception as e:
                print(folder, e)
                pass

    return database


def to_df(database_dict):
    df = pd.DataFrame()
    columns = ['patient_ID', 'study', 'study_description',  # noqa
                                       'series', 'series_description', 'subseries', 'modality', 
                                       'instances', 'instance_uid', 
                                       'reference_ct', 'reference_rs', 'reference_pl', 'reference_frame', 'folder',
                                       'orientation', 'orientation_type', 'MR_repetition_time', 'MR_echo_time', 
                                       'MR_scan_sequence', 'MR_magnetic_field_strength', 'MR_imaged_nucleus', 'file_path']
    for pat in database_dict:
        for study in database_dict[pat]:
            for series in database_dict[pat][study]:
                if series != "description":  # skip description key in dict
                    for subseries in database_dict[pat][study][series]:
                        if subseries != 'description':  # skip description key in dict
                            
                            values = [pat, study, database_dict[pat][study]['description'], # noqa
                                      series, database_dict[pat][study][series]['description'], 
                                      subseries, database_dict[pat][study][series][subseries]['modality'], 
                                      len(database_dict[pat][study][series][subseries]['instances']), database_dict[pat][study][series][subseries]['instance_uid'], 
                                      database_dict[pat][study][series][subseries]['reference_ct'], database_dict[pat][study][series][subseries]['reference_rs'], 
                                      database_dict[pat][study][series][subseries]['reference_pl'], database_dict[pat][study][series][subseries]['reference_frame'], database_dict[pat][study][series][subseries]['folder'],
                                      database_dict[pat][study][series][subseries]['orientation'], database_dict[pat][study][series][subseries]['orientation_type'],
                                      database_dict[pat][study][series][subseries]['repetition_time'], database_dict[pat][study][series][subseries]['echo_time'],
                                      database_dict[pat][study][series][subseries]['scan_sequence'], database_dict[pat][study][series][subseries]['mag_field_strength'], database_dict[pat][study][series][subseries]['imaged_nucleus'],
                                      database_dict[pat][study][series][subseries]['fname']
                                      ]

                            df_add = pd.DataFrame([values], columns=columns)
                            df = pd.concat([df, df_add], ignore_index=True)
    return df

# fmt: on


def crawl(
    top: pathlib.Path,
    n_jobs: int = -1,
    csv_path: pathlib.Path | None = None,
    json_path: pathlib.Path | None = None,
    imgtools_dir: str = ".imgtools",
) -> dict:
    database_list = []
    # folders = glob.glob(pathlib.Path(top, "*").as_posix())
    folders = [f for f in pathlib.Path(top).iterdir() if f.is_dir()]
    logger.info(f"Crawling {len(folders)} folders in {top}")

    database_list = Parallel(n_jobs=n_jobs)(
        delayed(crawl_one)(folder.as_posix())
        for folder in tqdm(folders, desc="Crawling folders", leave=False)
    )
    logger.info("Converting list to dictionary")
    # convert list to dictionary
    database_dict = {}
    for db in database_list:
        for key in db:
            database_dict[key] = db[key]

    # configure json path
    if json_path is None:
        json_path = top.parent / imgtools_dir / f"imgtools_{top.name}.json"

    json_path.parent.mkdir(parents=True, exist_ok=True)

    # configure csv path
    if csv_path is None:
        csv_path = top.parent / imgtools_dir / f"imgtools_{top.name}.csv"

    csv_path.parent.mkdir(parents=True, exist_ok=True)

    # Save as json
    logger.info(f"Saving as json to {json_path}")
    with json_path.open("w") as f:
        json.dump(database_dict, f, indent=4)

    # Save as dataframe
    logger.info(f"Saving as dataframe to {csv_path}")
    df = to_df(database_dict)
    df.to_csv(csv_path)

    return database_dict


if __name__ == "__main__":
    parser = ArgumentParser("Dataset DICOM Crawler")
    parser.add_argument(
        "directory", type=str, help="Top-level directory of the dataset."
    )
    parser.add_argument(
        "--n_jobs",
        type=int,
        default=os.cpu_count() - 2,
        help="Number of parallel processes for multiprocessing.",
    )

    args = parser.parse_args()
    db = crawl(args.directory, n_jobs=args.n_jobs)
    print("# patients:", len(db))



import copy
from typing import Dict, TypeVar

import pydicom

T = TypeVar("T")


def get_modality_metadata(dicom_data, modality: str):
    keys = {
        "ALL": {
            "BodyPartExamined": "BodyPartExamined",
            "DataCollectionDiameter": "DataCollectionDiameter",
            "NumberofSlices": "NumberofSlices",
            "SliceThickness": "SliceThickness",
            "ScanType": "ScanType",
            "ScanProgressionDirection": "ScanProgressionDirection",
            "PatientPosition": "PatientPosition",
            "ContrastType": "ContrastBolusAgent",
            "Manufacturer": "Manufacturer",
            "ScanOptions": "ScanOptions",
            "RescaleType": "RescaleType",
            "RescaleSlope": "RescaleSlope",
            "ManufacturerModelName": "ManufacturerModelName",
        },
        "CT": {
            "KVP": "KVP",
            "XRayTubeCurrent": "XRayTubeCurrent",
            "ScanOptions": "ScanOptions",
            "ReconstructionAlgorithm": "ReconstructionAlgorithm",
            "ContrastFlowRate": "ContrastFlowRate",
            "ContrastFlowDuration": "ContrastFlowDuration",
            "ContrastType": "ContrastBolusAgent",
            "ReconstructionMethod": "ReconstructionMethod",
            "ReconstructionDiameter": "ReconstructionDiameter",
            "ConvolutionKernel": "ConvolutionKernel",
        },
        "MR": [
            "AcquisitionTime",
            "AcquisitionContrast",
            "AcquisitionType",
            "RepetitionTime",
            "EchoTime",
            "ImagingFrequency",
            "MagneticFieldStrength",
            "SequenceName",
        ],
        "PT": [
            "RescaleType",
            "RescaleSlope",
            "RadionuclideTotalDose",
            "RadionuclideHalfLife",
        ],
    }

    # initialize metadata dictionary
    if modality == "ALL":
        metadata = {}
    else:
        metadata = all_modalities_metadata(dicom_data)

    # populating metadata
    if modality == "RTSTRUCT":
        if hasattr(dicom_data, "StructureSetROISequence"):
            metadata["numROIs"] = str(len(dicom_data.StructureSetROISequence))
    elif modality in keys:
        keys_mod = keys[modality]
        if isinstance(keys_mod, dict):
            for k in keys_mod:
                if hasattr(dicom_data, keys_mod[k]):
                    metadata[k] = getattr(dicom_data, keys_mod[k])
        elif isinstance(keys_mod, list):
            for k in keys_mod:
                if hasattr(dicom_data, k):
                    metadata[k] = getattr(dicom_data, k)
        else:
            print("WAGUANWAGUANWAGUAN")

    return metadata


def all_modalities_metadata(dicom_data: pydicom.dataset.FileDataset) -> Dict[str, T]:
    metadata = get_modality_metadata(dicom_data, "ALL")

    if hasattr(dicom_data, "PixelSpacing") and hasattr(dicom_data, "SliceThickness"):
        pixel_size = copy.copy(dicom_data.PixelSpacing)
        pixel_size.append(dicom_data.SliceThickness)
        metadata["PixelSize"] = str(tuple(pixel_size))

    return metadata



import warnings
from itertools import chain

from joblib import Parallel, delayed

from .ops import BaseOp, BaseInput, BaseOutput


class Pipeline:
    """Base class for image processing pipelines.

    A pipeline can be created by subclassing, instantiating the required
    data loaders and operations from `ops` in the constructor and implementing
    the `process_one_case` method, which defines the processing steps for one
    case (i.e. one subject_id from data loaders).
    """
    def __init__(self, n_jobs=1, missing_strategy="drop", show_progress=True, warn_on_error=False):
        """Initialize the base class.

        Parameters
        ----------
        n_jobs : int, optional
            The number of worker processes to use for parallel computation
            (default 0).
        """
        self.n_jobs = n_jobs
        self.missing_strategy = missing_strategy.lower()
        self.show_progress = show_progress
        self.warn_on_error = warn_on_error
        if self.missing_strategy not in ["drop", "pass"]:
            raise ValueError(f"missing_strategy must be either of 'drop' or 'pass', got {missing_strategy}")

    def _get_loader_subject_ids(self):
        loaders = (v._loader for v in self.__dict__.values() if isinstance(v, BaseInput))
        all_subject_ids = [loader.keys() for loader in loaders]
        unique_subject_ids = set(chain.from_iterable(all_subject_ids))

        if not all_subject_ids:
            raise AttributeError("Pipeline must define at least one input op (subclass of ops.BaseInput)")

        result = []
        for subject_id in unique_subject_ids:
            if not all((subject_id in subject_ids for subject_ids in all_subject_ids)):
                # TODO give more details about which input data is missing
                message = f"Subject {subject_id} is missing some of the input data "
                if self.missing_strategy == "drop":
                    message += f"and will be dropped according to current missing strategy ('{self.missing_strategy}')."
                elif self.missing_strategy == "pass":
                    message += f"but will be passed according to current missing strategy ('{self.missing_strategy}')."
                    result.append(subject_id)
                warnings.warn(message, category=RuntimeWarning)
                continue
            result.append(subject_id)

        return result

    @property
    def ops(self):
        # TODO (Michal) return ops in actual order of execution
        return [v for v in self.__dict__.values() if isinstance(v, BaseOp)]

    def __repr__(self):
        attrs = [(k, v) for k, v in self.__dict__.items() if not isinstance(v, BaseOp) and not k.startswith("_")]
        args = ", ".join(f"{k}={v}" for k, v in attrs)
        return f"{self.__class__.__module__}.{self.__class__.__name__}({args})"

    def __str__(self):
        repr_ = self.__repr__()
        ops = self.ops()
        if not ops:
            ops = "<none>"
        else:
            ops = ",\n".join(ops)
        return repr_ + "\n" + "ops: (\n" + ops + "\n)"

    def process_one_subject(self, subject_id):
        """Define the processing steps for one case.

        Parameters
        ----------
        subject_id : str
            The ID of the subject to be processed.
        """
        raise NotImplementedError

    def _process_wrapper(self, subject_id):
        try:
            print(subject_id)
            self.process_one_subject(subject_id)
        except Exception as e:
            message = f"{type(e).__name__} while processing subject {subject_id}: " + str(e)
            if self.warn_on_error:
                warnings.warn(message, category=RuntimeWarning)
            else:
                raise RuntimeError(message) from e

    def run(self):
        """Execute the pipeline, possibly in parallel.
        """
        # Joblib prints progress to stdout if verbose > 50
        verbose = 51 if self.show_progress else 0

        subject_ids = self._get_loader_subject_ids()
        # Note that returning any SimpleITK object in process_one_subject is
        # not supported yet, since they cannot be pickled
        Parallel(n_jobs=self.n_jobs, verbose=verbose)(
            delayed(self._process_wrapper)(subject_id) for subject_id in subject_ids)


class SequentialPipeline(Pipeline):
    def __init__(self, ops_list):
        self.source = ops_list.pop(0)
        self.sink = ops_list.pop() if isinstance(ops_list[-1], BaseOutput) else None
        self.ops_list = ops_list

    def process_one_subject(self, subject_id):
        image = self.source(subject_id)
        for op in self.ops_list:
            image = op(image)
        if self.sink is not None:
            self.sink(image)
        return image



import datetime
import tempfile
from pathlib import Path

import pytest
from pydicom.dataset import Dataset, FileMetaDataset
from pydicom.errors import InvalidDicomError
from pydicom.uid import UID, ExplicitVRLittleEndian

from imgtools.dicom.sort.utils import read_tags, sanitize_file_name, truncate_uid


@pytest.fixture(scope='module')
def dicom_test_file():
    """Pytest fixture to create a DICOM file for testing."""
    ds = Dataset()
    # Patient and instance-related metadata
    ds.PatientName = 'Test^Firstname'
    ds.PatientID = '123456'

    # Dates and times
    datetime.datetime.now()
    ds.ContentDate = '20021114'
    ds.ContentTime = '111131'  # Updated time format

    # Other metadata
    ds.SpecificCharacterSet = 'ISO_IR 100'
    ds.ImageType = ['ORIGINAL', 'PRIMARY', 'AXIAL']
    ds.StudyDate = '20021114'
    ds.SeriesDate = '20021114'
    ds.AcquisitionDate = '20021114'
    ds.StudyTime = '105444'
    ds.SeriesTime = '110039'
    ds.AcquisitionTime = '110234.982284'
    ds.Modality = 'CT'
    ds.Manufacturer = 'GE MEDICAL SYSTEMS'
    ds.StudyDescription = 'CT ABD & PELVIS W/O &'
    ds.SeriesDescription = '2.5SOFT + 30%ASIR'
    ds.ManufacturerModelName = 'LightSpeed VCT'
    ds.SeriesInstanceUID = UID('1.2.840.113619.2.55.3.604688.12345678.1234567890')
    ds.StudyInstanceUID = UID('1.2.840.113619.2.55.3.604688.98765432.9876543210')
    # File meta information
    file_meta = FileMetaDataset()
    file_meta.MediaStorageSOPClassUID = UID('1.2.840.10008.5.1.4.1.1.2')
    file_meta.MediaStorageSOPInstanceUID = UID(
        '1.3.6.1.4.1.14519.5.2.1.1706.4016.292639580813240923432069920621'
    )
    file_meta.ImplementationClassUID = UID('1.2.3.4')
    file_meta.TransferSyntaxUID = ExplicitVRLittleEndian

    ds.file_meta = file_meta

    # Create a temporary file
    temp_file = Path(tempfile.NamedTemporaryFile(suffix='.dcm', delete=False).name)
    ds.save_as(temp_file, write_like_original=False)

    yield temp_file

    # Cleanup
    temp_file.unlink()


class TestReadTags:
    def test_read_tags_with_main_metadata(self, dicom_test_file) -> None:
        tags = [
            'PatientName',
            'PatientID',
            'ContentDate',
            'ContentTime',
            'SpecificCharacterSet',
            'StudyDate',
            'SeriesDate',
            'AcquisitionDate',
            'StudyTime',
            'SeriesTime',
            'AcquisitionTime',
            'Modality',
            'Manufacturer',
            'StudyDescription',
            'SeriesDescription',
            'ManufacturerModelName',
        ]
        result = read_tags(file=dicom_test_file, tags=tags, truncate=True, sanitize=True)
        assert result['PatientName'] == sanitize_file_name('Test^Firstname')
        assert result['PatientID'] == sanitize_file_name('123456')
        assert result['ContentDate'] == sanitize_file_name('20021114')
        assert result['ContentTime'] == sanitize_file_name('111131')
        assert result['SpecificCharacterSet'] == sanitize_file_name('ISO_IR 100')
        assert result['StudyDate'] == sanitize_file_name('20021114')
        assert result['SeriesDate'] == sanitize_file_name('20021114')
        assert result['AcquisitionDate'] == sanitize_file_name('20021114')
        assert result['StudyTime'] == sanitize_file_name('105444')
        assert result['SeriesTime'] == sanitize_file_name('110039')
        assert result['AcquisitionTime'] == sanitize_file_name('110234.982284')
        assert result['Modality'] == sanitize_file_name('CT')
        assert result['Manufacturer'] == sanitize_file_name('GE MEDICAL SYSTEMS')
        assert result['StudyDescription'] == sanitize_file_name('CT ABD & PELVIS W/O &')
        assert result['SeriesDescription'] == sanitize_file_name('2.5SOFT + 30%ASIR')
        assert result['ManufacturerModelName'] == sanitize_file_name('LightSpeed VCT')

    def test_read_tags_with_empty_tags(self, dicom_test_file) -> None:
        tags = []
        result = read_tags(file=dicom_test_file, tags=tags, truncate=True, sanitize=True)
        assert result == {}

    def test_read_tags_with_nonexistent_tags(self, dicom_test_file) -> None:
        tags = ['NonexistentTag']
        with pytest.raises(ValueError):
            read_tags(file=dicom_test_file, tags=tags, truncate=True, sanitize=True)

    def test_read_tags_with_partial_metadata(self, dicom_test_file) -> None:
        tags = ['PatientName', 'Modality', 'NonexistentTag']
        with pytest.raises(ValueError):
            read_tags(file=dicom_test_file, tags=tags, truncate=True, sanitize=True)
        # assert result['PatientName'] == 'Test^Firstname'
        # assert result['Modality'] == 'CT'
        # assert 'NonexistentTag' not in result

    def test_read_tags_with_truncate_false(self, dicom_test_file) -> None:
        tags = ['SeriesInstanceUID']
        result = read_tags(file=dicom_test_file, tags=tags, truncate=False, sanitize=True)
        assert result['SeriesInstanceUID'] == '1.2.840.113619.2.55.3.604688.12345678.1234567890'

    def test_read_tags_with_sanitize_false(self, dicom_test_file) -> None:
        tags = ['StudyDescription']
        result = read_tags(file=dicom_test_file, tags=tags, truncate=True, sanitize=False)
        assert result['StudyDescription'] == 'CT ABD & PELVIS W/O &'

    def test_none_file(self) -> None:
        with pytest.raises(AssertionError):
            read_tags(file=None, tags=['PatientID'], truncate=True, sanitize=True)

    def test_invalid_dicom_file(self) -> None:
        inv_file = Path('invalid.dcm')
        inv_file.touch()
        with pytest.raises(InvalidDicomError):
            read_tags(file=Path(inv_file), tags=['PatientID'], truncate=True, sanitize=True)


class TestTruncateUid:
    def test_truncate_uid_with_default_last_digits(self) -> None:
        uid = '1.2.840.10008.1.2.1'
        result = truncate_uid(uid)
        assert result == '1.2.1'

    def test_truncate_uid_with_custom_last_digits(self) -> None:
        uid = '1.2.840.10008.1.2.1'
        result = truncate_uid(uid, last_digits=10)
        assert result == '0008.1.2.1'

    def test_truncate_uid_with_short_uid(self) -> None:
        uid = '12345'
        result = truncate_uid(uid, last_digits=10)
        assert result == '12345'

    def test_truncate_uid_with_zero_last_digits(self) -> None:
        uid = '1.2.840.10008.1.2.1'
        result = truncate_uid(uid, last_digits=0)
        assert result == '1.2.840.10008.1.2.1'

    def test_truncate_uid_with_negative_last_digits(self) -> None:
        uid = '1.2.840.10008.1.2.1'
        result = truncate_uid(uid, last_digits=-5)
        assert result == '1.2.840.10008.1.2.1'



from .datagraph import DataGraph
from .dose import Dose
from .pet import PET
from .scan import Scan
from .segmentation import Segmentation, accepts_segmentations, map_over_labels
from .sparsemask import SparseMask
from .structureset import StructureSet

__all__ = [
    "Segmentation",
    "map_over_labels",
    "accepts_segmentations",
    "StructureSet",
    "PET",
    "Dose",
    "DataGraph",
    "SparseMask",
    "Scan",
]



"""DICOM Sorting Utilities.

This module provides utilities for sanitizing filenames, truncating UIDs,
and reading specific tags from DICOM files.

Functions
---------
sanitize_file_name(filename: str) -> str
    Sanitize filenames by replacing potentially dangerous characters.
truncate_uid(uid: str, last_digits: int = 5) -> str
    Truncate a UID to the last n characters.
read_tags(file: Path, tags: list[str], truncate: bool = True, sanitize: bool = True) -> dict[str, str]
    Read specified tags from a DICOM file.

Examples
--------
Sanitize a filename:
    >>> sanitize_file_name("test<>file:/name.dcm")
    'test_file_name.dcm'

Truncate a UID:
    >>> truncate_uid(
    ...     "1.2.840.10008.1.2.1",
    ...     last_digits=5,
    ... )
    '.1.2.1'

Read tags from a DICOM file:
    >>> from pathlib import (
    ...     Path,
    ... )
    >>> tags = [
    ...     "PatientID",
    ...     "StudyInstanceUID",
    ... ]
    >>> read_tags(
    ...     Path("sample.dcm"),
    ...     tags,
    ... )
    {'PatientID': '12345', 'StudyInstanceUID': '1.2.3.4.5'}
"""

import re
from pathlib import Path
from typing import Dict, List, Optional

from pydicom import dcmread
from pydicom.errors import InvalidDicomError


def sanitize_file_name(filename: str) -> str:
    """
    Sanitize the file name by replacing potentially dangerous characters.

    Replaces disallowed characters with underscores, converts spaces to
    underscores, and removes subsequent underscores.

    Parameters
    ----------
    filename : str
        The input file name to sanitize.

    Returns
    -------
    str
        The sanitized file name.

    Examples
    --------
    >>> sanitize_file_name("test<>file:/name.dcm")
    'test_file_name.dcm'
    >>> sanitize_file_name("my file name.dcm")
    'my_file_name.dcm'
    """
    assert filename is not None
    assert isinstance(filename, str)

    disallowed_characters_pattern = re.compile(r'[<>:"/\\|?*\x00-\x1f]')
    sanitized_name = disallowed_characters_pattern.sub("_", filename)
    sanitized_name = sanitized_name.replace(" ", "_")
    sanitized_name = re.sub(r"(_{2,})", "_", sanitized_name)
    return sanitized_name


def truncate_uid(uid: str, last_digits: int = 5) -> str:
    """
    Truncate the UID to the last n characters (including periods and underscores).

    If the UID is shorter than `last_digits`, the entire UID is returned.

    Parameters
    ----------
    uid : str
        The UID string to truncate.
    last_digits : int, optional
        The number of characters to keep at the end of the UID (default is 5).

    Returns
    -------
    str
        The truncated UID string.

    Examples
    --------
    >>> truncate_uid(
    ...     "1.2.840.10008.1.2.1",
    ...     last_digits=5,
    ... )
    '.1.2.1'
    >>> truncate_uid(
    ...     "12345",
    ...     last_digits=10,
    ... )
    '12345'
    """
    assert uid is not None
    assert isinstance(uid, str)
    assert isinstance(last_digits, int)
    if last_digits >= len(uid) or last_digits <= 0:
        return uid

    return uid[-last_digits:]


def read_tags(
    file: Path,
    tags: List[str],
    truncate: bool = True,
    sanitize: bool = True,
    default: Optional[str] = "",
) -> Dict[str, str]:
    """
    Read the specified tags from a DICOM file.

    Reads a set of tags from a DICOM file and applies optional sanitization
    and truncation for UIDs. Handles cases where specific tags may be missing.

    Parameters
    ----------
    file : Path
                                    Path to the DICOM file.
    tags : list of str
                                    List of DICOM tags to read.
    truncate : bool, optional
                                    If True, truncate UIDs to the last 5 characters (default is True).
    sanitize : bool, optional
                                    If True, sanitize tag values to remove disallowed characters
                                    (default is True).

    Returns
    -------
    dict of str : str
                                    A dictionary mapping tags to their values.

    Raises
    ------
    TypeError
                                    If there is a type error while reading the DICOM file.
    InvalidDicomError
                                    If the file is not a valid DICOM file.
    ValueError
                                    If there is a value error while reading the file.

    Notes
    -----
    For RTSTRUCT files, missing 'InstanceNumber' tags default to '1'.

    Examples
    --------
    Read tags from a valid DICOM file with truncation:
    >>> from pathlib import (
    ...     Path,
    ... )
    >>> read_tags(
    ...     Path("sample.dcm"),
    ...     [
    ...         "PatientID",
    ...         "StudyInstanceUID",
    ...     ],
    ... )
    {'PatientID': '12345', 'StudyInstanceUID': '1.2.3.4.5'}

    Read tags without truncating UIDs:
    >>> read_tags(
    ...     Path("sample.dcm"),
    ...     [
    ...         "PatientID",
    ...         "StudyInstanceUID",
    ...     ],
    ...     truncate=False,
    ... )
    {'PatientID': '12345', 'StudyInstanceUID': '1.2.840.10008.1.2.1'}

    Handle missing tags:
    >>> read_tags(
    ...     Path("sample.dcm"),
    ...     ["NonexistentTag"],
    ... )
    [warn] No value for tag: NonexistentTag in file: sample.dcm
    {'NonexistentTag': 'UNKNOWN'}
    """
    assert isinstance(file, Path)
    assert isinstance(tags, list) and all(isinstance(tag, str) for tag in tags) and tags is not None

    try:
        dicom = dcmread(file, specific_tags=tags, stop_before_pixels=True)
    except FileNotFoundError as fnfe:
        errmsg = f"File not found: {file}"
        raise FileNotFoundError(errmsg) from fnfe
    except InvalidDicomError as ide:
        errmsg = f"Invalid DICOM file: {file}"
        raise InvalidDicomError(errmsg) from ide
    except ValueError as ve:
        errmsg = f"Value error reading DICOM file: {file}"
        raise ValueError(errmsg) from ve

    result = {}
    # is_rtstruct = dicom.get('Modality') == 'RTSTRUCT' if 'InstanceNumber' in tags else False

    for tag in tags:
        value = str(dicom.get(tag, default=default))

        if truncate and tag.endswith("UID"):
            value = truncate_uid(value)

        if sanitize:
            value = sanitize_file_name(value)

        result[tag] = value
    return result



import csv
import json
import os
import pathlib
import pickle
import shutil
from datetime import datetime, timezone

import h5py
import nrrd
import numpy as np
import SimpleITK as sitk
from skimage.measure import regionprops

from imgtools.utils import image_to_array


class BaseWriter:
    def __init__(self, root_directory, filename_format, create_dirs=True):
        self.root_directory = root_directory
        self.filename_format = filename_format
        self.create_dirs = create_dirs
        if create_dirs and not os.path.exists(self.root_directory):
            os.makedirs(self.root_directory)

    def put(self, *args, **kwargs):
        raise NotImplementedError

    def _get_path_from_subject_id(self, subject_id, **kwargs):
        now = datetime.now(timezone.utc)
        date = now.strftime("%Y-%m-%d")
        time = now.strftime("%H%M%S")
        date_time = date + "_" + time
        out_filename = self.filename_format.format(
            subject_id=subject_id, date=date, time=time, date_time=date_time, **kwargs
        )
        out_path = pathlib.Path(self.root_directory, out_filename).as_posix()
        out_dir = os.path.dirname(out_path)
        if self.create_dirs and not os.path.exists(out_dir):
            os.makedirs(
                out_dir, exist_ok=True
            )  # create subdirectories if specified in filename_format

        return out_path


class BaseSubjectWriter(BaseWriter):
    def __init__(
        self,
        root_directory,
        filename_format="{subject_id}.nii.gz",
        create_dirs=True,
        compress=True,
    ):
        super().__init__(root_directory, filename_format, create_dirs)
        self.root_directory = root_directory
        self.filename_format = filename_format
        self.create_dirs = create_dirs
        self.compress = compress
        if os.path.exists(self.root_directory):
            # delete the folder called {subject_id} that was made in the original BaseWriter / the one named {label_or_image}
            if os.path.basename(os.path.dirname(self.root_directory)) == "{subject_id}":
                shutil.rmtree(os.path.dirname(self.root_directory))
            elif "{label_or_image}{train_or_test}" in os.path.basename(
                self.root_directory
            ):
                shutil.rmtree(self.root_directory)

    def put(
        self,
        subject_id,
        image,
        is_mask=False,
        nnunet_info=None,
        label_or_image: str = "images",
        mask_label: str = "",
        train_or_test: str = "Tr",
        **kwargs,
    ):
        if is_mask:
            # remove illegal characters for Windows/Unix
            badboys = r'<>:"/\|?*'
            for char in badboys:
                mask_label = mask_label.replace(char, "")

            # filename_format eh
            self.filename_format = (
                mask_label + ".nii.gz"
            )  # save the mask labels as their rtstruct names

        if nnunet_info:
            if label_or_image == "labels":
                filename = f"{subject_id}.nii.gz"  # naming convention for labels
            else:
                filename = self.filename_format.format(
                    subject_id=subject_id,
                    modality_index=nnunet_info["modalities"][
                        nnunet_info["current_modality"]
                    ],
                )  # naming convention for images
            out_path = self._get_path_from_subject_id(
                filename, label_or_image=label_or_image, train_or_test=train_or_test
            )
        else:
            out_path = self._get_path_from_subject_id(
                self.filename_format, subject_id=subject_id
            )
        sitk.WriteImage(image, out_path, self.compress)

    def _get_path_from_subject_id(self, filename, **kwargs):
        root_directory = self.root_directory.format(
            **kwargs
        )  # replace the {} with the kwargs passed in from .put() (above)
        out_path = pathlib.Path(root_directory, filename).as_posix()
        out_dir = os.path.dirname(out_path)
        if self.create_dirs and not os.path.exists(out_dir):
            os.makedirs(
                out_dir, exist_ok=True
            )  # create subdirectories if specified in filename_format
        return out_path


class ImageFileWriter(BaseWriter):
    def __init__(
        self,
        root_directory,
        filename_format="{subject_id}.nii.gz",
        create_dirs=True,
        compress=True,
    ):
        super().__init__(root_directory, filename_format, create_dirs)
        self.compress = compress

    def put(self, subject_id, image, **kwargs):
        out_path = self._get_path_from_subject_id(subject_id, **kwargs)
        sitk.WriteImage(image, out_path, self.compress)


class SegNrrdWriter(BaseWriter):
    def __init__(
        self,
        root_directory,
        filename_format="{subject_id}.seg.nrrd",
        create_dirs=True,
        compress=True,
    ):
        super().__init__(root_directory, filename_format, create_dirs)
        if compress:
            self.compression_level = 9
        else:
            self.compression_level = 1

    def put(self, subject_id, mask, **kwargs):
        out_path = self._get_path_from_subject_id(subject_id, **kwargs)
        labels = [k for k in mask.roi_names]
        print(labels)

        origin = mask.GetOrigin()
        spacing = mask.GetSpacing()
        # direction = mask.GetDirection()

        space = "left-posterior-superior"  # everything is ITK read/write

        # fix reverted somewhere.... :''''(
        space_directions = [
            [spacing[0], 0.0, 0.0],
            [0.0, spacing[1], 0.0],
            [0.0, 0.0, spacing[2]],
        ]
        kinds = ["domain", "domain", "domain"]
        dims = 3

        # permute axes to original orientations
        if len(labels) > 1:
            arr = np.transpose(sitk.GetArrayFromImage(mask), [-1, -2, -3, -4])

            # add extra dimension to metadata
            space_directions.insert(0, [float("nan"), float("nan"), float("nan")])
            kinds.insert(0, "vector")
            dims += 1
        else:
            arr = np.transpose(sitk.GetArrayFromImage(mask), [-1, -2, -3])

        # ensure proper conversion to array
        assert mask.GetSize() == arr.shape[-3:]

        segment_info = {}
        for n, i in enumerate(labels):
            try:
                if len(labels) > 1:
                    props = regionprops(arr[n])[0]
                else:
                    props = regionprops(arr)[0]
                bbox = props["bbox"]
                bbox_segment = [bbox[0], bbox[3], bbox[1], bbox[4], bbox[2], bbox[5]]
            except IndexError:  # mask is empty
                assert (
                    arr[n].sum() == 0
                ), "Mask not empty but 'skimage.measure.regionprops' failed."
                bbox_segment = [0, 0, 0, 0, 0, 0]

            segment_info[f"Segment{n}_Color"] = list(np.random.random(3))
            segment_info[f"Segment{n}_ColorAutoGenerated"] = "1"
            segment_info[f"Segment{n}_Extent"] = bbox_segment
            segment_info[f"Segment{n}_ID"] = str(n)
            segment_info[f"Segment{n}_Name"] = i
            segment_info[f"Segment{n}_NameautoGenerated"] = "0"

        header = {
            "dimension": dims,
            "space": space,
            "sizes": mask.GetSize(),
            "space directions": space_directions,
            "kinds": kinds,
            "endian": "little",
            "space origin": origin,
            "roi_names": labels,
            **segment_info,
        }

        nrrd.write(
            out_path,
            arr,
            header=header,
            compression_level=self.compression_level,
            **kwargs,
        )


class NumpyWriter(BaseWriter):
    def __init__(
        self, root_directory, filename_format="{subject_id}.npy", create_dirs=True
    ):
        super().__init__(root_directory, filename_format, create_dirs)
        self.root_directory = root_directory
        self.filename_format = filename_format

    def put(self, subject_id, image, **kwargs):
        out_path = self._get_path_from_subject_id(subject_id, **kwargs)
        if isinstance(image, sitk.Image):
            array, *_ = image_to_array(
                image
            )  # TODO (Michal) optionally save the image geometry
        np.save(out_path, array)


class HDF5Writer(BaseWriter):
    def __init__(
        self,
        root_directory,
        filename_format="{subject_id}.h5",
        create_dirs=True,
        save_geometry=True,
    ):
        super().__init__(root_directory, filename_format, create_dirs)
        self.save_geometry = save_geometry

    def put(self, subject_id, images, metadata=None, **kwargs):
        out_path = self._get_path_from_subject_id(subject_id, **kwargs)
        with h5py.File(out_path, "w") as f:
            if not isinstance(images, dict):
                images = {"image": images}
            for k, v in images.items():
                array, origin, direction, spacing = image_to_array(v)
                dataset = f.create_dataset(k, data=array)
                dataset.attrs.create("subject_id", subject_id)
                if self.save_geometry:
                    dataset.attrs.create("origin", data=origin)
                    dataset.attrs.create("direction", data=direction)
                    dataset.attrs.create("spacing", data=spacing)
            if metadata:
                for k, attrs in metadata.items():
                    for name, v in attrs:
                        f[subject_id].attrs.create(name, data=v)


class MetadataWriter(BaseWriter):
    def __init__(
        self,
        root_directory,
        filename_format="{subject_id}.json",
        create_dirs=True,
        remove_existing=True,
    ):
        super().__init__(root_directory, filename_format, create_dirs)
        self.file_format = os.path.splitext(filename_format)[1].lstrip(".")
        self.remove_existing = remove_existing
        if self.file_format not in ["json", "csv", "pkl"]:
            raise ValueError(
                f"File format {self.file_format} not supported. Supported formats: JSON (.json), CSV (.csv), Pickle (.pkl)."
            )

        if self.file_format == "csv" and self.remove_existing:
            out_path = pathlib.Path(
                self.root_directory, self.filename_format
            ).as_posix()
            if os.path.exists(out_path):
                os.remove(out_path)  # remove existing CSV instead of appending

    def _put_json(self, out_path, **kwargs):
        with open(out_path, "w") as f:
            json.dump(kwargs, f)

    def _put_csv(self, out_path, **kwargs):
        with open(out_path, "a+") as f:
            writer = csv.DictWriter(f, fieldnames=kwargs.keys())
            pos = f.tell()
            f.seek(0)
            sample = "\n".join([f.readline() for _ in range(2)])
            if sample == "\n" or not csv.Sniffer().has_header(sample):
                writer.writeheader()
            f.seek(pos)
            writer.writerow(kwargs)

    def _put_pickle(self, out_path, **kwargs):
        with open(out_path, "wb") as f:
            pickle.dump(kwargs, f)

    def put(self, subject_id, **kwargs):
        out_path = self._get_path_from_subject_id(subject_id)

        if "subject_id" not in kwargs:
            kwargs["subject_id"] = subject_id

        if self.file_format == "json":
            self._put_json(out_path, **kwargs)
        elif self.file_format == "csv":
            self._put_csv(out_path, **kwargs)
        elif self.file_format == "pkl":
            self._put_pickle(out_path, **kwargs)
